"""
Default ConfigUI for S4-E25D40 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dWOkAoRjBQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABUCZUAWVQ3'
    b'VzdG9tIFJlY29yZGluZyBUYWdzUAGDF/9/UQSAUSSBQFAV8UtleXdvcmRzIGZvciBvcmdh'
    b'bml6aW5nIGFuZCBpZGVudGlmeWluZyByZWNvcmRpbmdzLCBzZXBhcmF0ZWQgYnkgY29tbW'
    b'Egb3Igc3BhY2UuIExpbWl0IGlzIDIwIHRhZ3MsIDY0IGNoYXJhY3RlcnMuQAX7UAWTUmVj'
    b'b3JkaW5nIERpcmVjdG9yeVABgxT/f1EEhlJFQ09SRFEkgRBQFJNDb25maWdbMHgxNkZGN0'
    b'ZdPT0wUBW5RGlyZWN0b3J5IGZvciBzZXJpYWxpemVkIHJlY29yZGluZ3MuIDE2IGNoYXJh'
    b'Y3RlciBtYXhpbXVtQAVAo1AFlVJlY29yZGluZyBGaWxlIFByZWZpeFABgxX/f1EkgRBQFe'
    b'hQcmVmaXggZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5nIG5hbWVzLiBJZiBibGFuaywgZGV2'
    b'aWNlIHNlcmlhbCBudW1iZXIgd2lsbCBiZSB1c2VkLiAxNiBjaGFyYWN0ZXIgbWF4aW11bV'
    b'AUk0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGB'
    b'AVAVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIH'
    b'RoZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1p'
    b'ZG5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZW'
    b'N0b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51'
    b'bWJlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBY'
    b'pUaW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQKlQBYtCdXR0b24gTW9kZVABgxD/f1EB'
    b'gQBQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3'
    b'IgYmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBT'
    b'ZWNvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BBB41QBYpTdGFydC1Pbmx5QBdBFFAFjlBsdW'
    b'ctSW4gQWN0aW9uUAGDCv9/UQGBAVAVrVNldCBhY3Rpb24gZm9yIGRldmljZSB3aGVuIHBs'
    b'dWdnZWQgaW4gdG8gVVNCLkEHtVAFskltbWVkaWF0ZWx5IHN0b3AgcmVjb3JkaW5nIGFuZC'
    b'BhcHBlYXIgYXMgVVNCIGRyaXZlQQezUAWwQ29tcGxldGUgcmVjb3JkaW5nIGJlZm9yZSBh'
    b'cHBlYXJpbmcgYXMgVVNCIGRyaXZlQQewUAWtSWdub3JlOiBzdG9wIHJlY29yZGluZyB3aG'
    b'VuIGJ1dHRvbiBpcyBwcmVzc2VkQQelUAWiU3RhcnQgcmVjb3JkaW5nIHdoZW4gVVNCIGNv'
    b'bm5lY3RlZEAj51ABgwv/f1AV3kVudGVyIGxvY2FsIHRpbWV6b25lJ3Mgb2Zmc2V0IGZyb2'
    b'0gVVRDIHRpbWUuIEltcGFjdHMgdGltZXN0YW1wIGluZm9ybWF0aW9uIG9uIHJlY29yZGlu'
    b'ZyBmaWxlcy5AL4BAD4BAT4BAKEjnUAWIVHJpZ2dlcnNAB0DaUAWMVHJpZ2dlciBNb2RlUA'
    b'GDEv9/UQGBAFAVQI9EZWxheSBUaGVuIFRyaWdnZXI6IFdhaXQgc3BlY2lmaWVkIHRpbWUs'
    b'IHRoZW4gd2FpdCBmb3IgdHJpZ2dlcgpEZWxheSBPciBUcmlnZ2VyOiBBY3F1aXJlIGFmdG'
    b'VyIHdhaXQgb3Igb24gdHJpZ2dlci4gQmF0dGVyeSBsaWZlIG1heSBiZSByZWR1Y2VkLkEH'
    b'lVAFkkRlbGF5IFRoZW4gVHJpZ2dlckEHk1AFkERlbGF5IE9yIFRyaWdnZXJAMvRQBY1TdG'
    b'FydCBhdCBUaW1lUAGDD/9/UBXbV2FpdCB1bnRpbCBzZXQgdGltZSBiZWZvcmUgZXZhbHVh'
    b'dGluZyBvdGhlciB0cmlnZ2VyIGNvbmRpdGlvbnMsIGluY2x1ZGluZyBhIHJlY29yZGluZy'
    b'BkZWxheUARrlAFlFJlY29yZGluZyBUaW1lIExpbWl0UAGDDf9/USGE////8FAlh1NlY29u'
    b'ZHNAEUCQUAWYUmVjb3JkaW5nIERlbGF5L0ludGVydmFsUAGDDP9/UQGBAFAlh1NlY29uZH'
    b'NRIYMBUYBQFdhUaW1lIGJlZm9yZSBzdGFydGluZyBhIHJlY29yZGluZywgYW5kIHRpbWUg'
    b'YmV0d2VlbiByZWNvcmRpbmdzIGlmICdSZXRyaWdnZXInIGlzIGNoZWNrZWQuQBGvUAWZUm'
    b'Vjb3JkaW5nIEZpbGUgU2l6ZSBMaW1pdFABgxH/f1EhhP////BQJYNLaUJAEECXUAWJUmV0'
    b'cmlnZ2VyUAGDDv9/UBSwQ29uZmlnWzB4REZGN0ZdPT1udWxsIGFuZCBDb25maWdbMHgxMU'
    b'ZGN0ZdPT1udWxsUBXPV2hlbiBzZXQsIFNsYW0gU3RpY2sgd2lsbCByZS1hcm0gYWZ0ZXIg'
    b'cmVjb3JkaW5nIHRpbWUgb3Igc2l6ZSBsaW1pdCBpcyByZWFjaGVkLkAQQKhQBZ5XYWl0IG'
    b'ZvciBBbGwgU2Vuc29yIENvbmRpdGlvbnNQAYMT/39QFf5XaGVuIHNldCwgcmVjb3JkaW5n'
    b'IHdpbGwgb25seSBzdGFydCB3aGVuIGFsbCB0cmlnZ2VycyBhcmUgbWV0LCBvdGhlcndpc2'
    b'UgcmVjb3JkaW5nIHdpbGwgc3RhcnQgd2hlbiBhbnkgc2Vuc29yIGNvbmRpdGlvbiBpcyBt'
    b'ZXRAGEDuUAWQUHJlc3N1cmUgVHJpZ2dlclABgwUAJFEAgQBAguVQBZVQcmVzc3VyZSBUcm'
    b'lnZ2VyLCBMb3dQAYMDACRRAYMBX5BQJYJQYVEhgwHUwFAVs1NldCB0byBzdGFydCBzYW1w'
    b'bGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cmUgd2luZG93LkCC5lAFllByZXNzdXJlIFRyaW'
    b'dnZXIsIEhpZ2hQAYMEACRRAYMBrbBQJYJQYVEhgwHUwFAVs1NldCB0byBzdGFydCBzYW1w'
    b'bGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cmUgd2luZG93LkAYQPtQBZNUZW1wZXJhdHVyZS'
    b'BUcmlnZ2VyUAGDBQEkUQCBAEBj6lAFmFRlbXBlcmF0dXJlIFRyaWdnZXIsIExvd1ABgwMB'
    b'JFECgfZQJYFDURKB2FEigVBQFbZTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaW'
    b'RlIHRlbXBlcmF0dXJlIHdpbmRvdy5AY+tQBZlUZW1wZXJhdHVyZSBUcmlnZ2VyLCBIaWdo'
    b'UAGDBAEkUQKBI1AlgUNREoHYUSKBUFAVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG'
    b'91dHNpZGUgdGVtcGVyYXR1cmUgd2luZG93LkAYQadQBZpBY2NlbGVyYXRpb24gVHJpZ2dl'
    b'ciAoMjVnKVABgwX/CFARgwf/IFEAgQBQJIBQFJVDb25maWdbMHgwMUZGMDhdID09IDBAQ0'
    b'CrUROIwDkAAAAAAABRI4hAOQAAAAAAAFAFmUFjY2VsZXJhdGlvbiBUcmlnZ2VyLCBMb3dQ'
    b'AYMD/whRAoL5mlAV4EhpZ2ggcG93ZXIgdHJpZ2dlciB3aXRoIHByZS10cmlnZ2VyIGRhdG'
    b'EuIE5vdCByZWNvbW1lbmRlZCBmb3IgbG9uZyB3YWl0cyB3aXRob3V0IGV4dGVybmFsIHBv'
    b'd2VyLlIDiD9JABkAGQAZQENArFETiMA5AAAAAAAAUSOIQDkAAAAAAABQBZpBY2NlbGVyYX'
    b'Rpb24gVHJpZ2dlciwgSGlnaFABgwT/CFECggZmUBXgSGlnaCBwb3dlciB0cmlnZ2VyIHdp'
    b'dGggcHJlLXRyaWdnZXIgZGF0YS4gTm90IHJlY29tbWVuZGVkIGZvciBsb25nIHdhaXRzIH'
    b'dpdGhvdXQgZXh0ZXJuYWwgcG93ZXIuUgOIP0kAGQAZABlAGEGjUAWaQWNjZWxlcmF0aW9u'
    b'IFRyaWdnZXIgKDQwZylQAYMH/1BQFdRMb3cgcG93ZXIgdHJpZ2dlciB1c2luZyBkaWdpdG'
    b'FsIGFjY2VsZXJvbWV0ZXIgd2l0aCBEQyByZXNwb25zZS4gTm8gcHJlLXRyaWdnZXIgZGF0'
    b'YS5RAIEAUBGDBf8IUBGDB/8gUCSAUBSUQ29uZmlnWzB4MWZmNTBdID09IDBAQ0CyUAWWQW'
    b'NjZWxlcmF0aW9uIFRocmVzaG9sZFABgwT/UFEBgiAAUgOIP0QAFAAUABRRI4hARAAAAAAA'
    b'AFETgFAV8kxvdyBwb3dlciB0cmlnZ2VyLiBUaGUgbWluaW11bSBhY2NlbGVyYXRpb24gKH'
    b'Bvc2l0aXZlIG9yIG5lZ2F0aXZlKSB0byB0cmlnZ2VyIGEgcmVjb3JkaW5nLiBNdXN0IGJl'
    b'IGdyZWF0ZXIgdGhhbiAwLkAnxlABgwX/UFEBgQdBB5FQBY5YIEF4aXMgVHJpZ2dlckEHkV'
    b'AFjlkgQXhpcyBUcmlnZ2VyQQeRUAWOWiBBeGlzIFRyaWdnZXJAD4BAT4BAKEcOUAWMTWVh'
    b'c3VyZW1lbnRzQAhCPFAFlzI1ZyBBY2NlbGVyYXRpb24gKENoIDgpUCSAQCf+UQGBB1ABgw'
    b'H/CFAVoEVuYWJsZS9EaXNhYmxlIGFjY2VsZXJhdGlvbiBheGVzQQeYUAWRRW5hYmxlIFgg'
    b'KENoIDguMClRAYEAQQeYUAWRRW5hYmxlIFkgKENoIDguMSlRAYEBQQeYUAWRRW5hYmxlIF'
    b'ogKENoIDguMilRAYECQBFAmlAFi1NhbXBsZSBSYXRlUAGDAv8IUBXYU2V0IG1haW4gYWNj'
    b'ZWxlcmF0aW9uIHNhbXBsZSByYXRlIGZyb20gMTAgdG8gMjAwMDAgSHouCkhpZ2hlciByYX'
    b'RlIHJlZHVjZXMgYmF0dGVyeSBsaWZlLlAUlUNvbmZpZ1sweDAxRkYwOF0gPT0gMFEBghOI'
    b'UCWCSHpREYEKUSGCTiBAEUD8UAWjT3ZlcnJpZGUgQW50aWFsaWFzaW5nIEZpbHRlciBDdX'
    b'RvZmZQAYMI/whRAYID6FAlgkh6USGCYahQFUClRGVmYXVsdCwgd2hlbiBub3QgY2hlY2tl'
    b'ZCwgb2Ygb25lLWZpZnRoIHNhbXBsZSBmcmVxdWVuY3kuIEN1dC1vZmYgZnJlcXVlbmN5IG'
    b'hhcyAyOSUgYXR0ZW51YXRpb24uCjV0aCBvcmRlciBCdXR0ZXJ3b3J0aC4gU29tZSBhdHRl'
    b'bnVhdGlvbiBhdCA2MCUgb2YgY3V0LW9mZiBmcmVxdWVuY3kuUBSVQ29uZmlnWzB4MDFGRj'
    b'A4XSA9PSAwQAhBHFAFmzQwZyBEQyBBY2NlbGVyYXRpb24gKENoIDgwKUAnn1ABgwH/UFEB'
    b'gQFBB5JQBY9FbmFibGUgQWxsIEF4ZXNAF0DYUAWLU2FtcGxlIFJhdGVQAYMC/1BRAYIB9F'
    b'Algkh6UBXeTG93IHBhc3MgZmlsdGVyIGlzIHNldCB0byAyNSUgb2YgZGF0YSByYXRlLCBz'
    b'byBhIGRhdGEgcmF0ZSBvZiAyMDAwSHogaGFzIGEgTFAgZmlsdGVyIGF0IDUwMCBIelAUk2'
    b'5vdCBDb25maWdbMHgxZmY1MF1BB4VRAYIPoEEHhVEBggfQQQeFUQGCA+hBB4VRAYIB9EEH'
    b'hFEBgfpBB4RRAYF9QQeEUQGBP0EHhFEBgSBBB4RRAYEQQAhBMFAFqENvbnRyb2wgUGFkIF'
    b'RlbXBlcmF0dXJlL1ByZXNzdXJlIChDaCA1OSlAJ0CbUQGBA1ABgwH/O1AVpUVuYWJsZS9E'
    b'aXNhYmxlIHNlbnNvcnMgb24gY29udHJvbCBwYWRBB5xQBZlFbmFibGUgUHJlc3N1cmUgKE'
    b'NoIDU5LjApQQefUAWcRW5hYmxlIFRlbXBlcmF0dXJlIChDaCA1OS4xKUEHpVAFokVuYWJs'
    b'ZSBSZWxhdGl2ZSBIdW1pZGl0eSAoQ2ggNTkuMilAEeNQBYtTYW1wbGUgUmF0ZVABgwL/O1'
    b'AUlENvbmZpZ1sweDFmZjNCXSA9PSAwUBWkU2V0IHNhbXBsZSBmcmVxdWVuY3kgZnJvbSAx'
    b'IHRvIDEwIEh6UQGBClAlgkh6URGBAVEhgQpACEHNUAWZSW5lcnRpYWwgTWVhc3VyZW1lbn'
    b'QgVW5pdEAHQJZQBZBBY3F1aXNpdGlvbiBNb2RlUAGDAf8rUCWBIFEBgRBBB4pQBYNPZmZR'
    b'AYEAQQeoUAWhQWJzb2x1dGUgT3JpZW50YXRpb24gKFF1YXRlcm5pb24pUQGBCEEHqFAFoV'
    b'JlbGF0aXZlIE9yaWVudGF0aW9uIChRdWF0ZXJuaW9uKVEBgRBBB49QBYhSb3RhdGlvblEB'
    b'gQJAAUCGUAWVT3JpZW50YXRpb24gRGF0YSBSYXRlUAGDAggrUQGBZFAlgkh6URGBAVEhgc'
    b'hQFb1TYW1wbGUgcmF0ZSBmb3IgZGlyZWN0aW9uYWwgZGF0YSBmcm9tIDEgdG8gMjAwIEh6'
    b'LiBDaGFubmVsIDY1UBSUQ29uZmlnWzB4MDFGRjJCXSA8IDhAAUCJUAWSUm90YXRpb24gRG'
    b'F0YSBSYXRlUAGDAgIrUQGBZFAlgkh6URGBAVEhgchQFb5TZXQgc2FtcGxlIHJhdGUgZm9y'
    b'IHJvdGF0aW9uIGRhdGEgZnJvbSAxIHRvIDIwMCBIei4gQ2hhbm5lbCA0N1AUmUNvbmZpZ1'
    b'sweDAxRkYyQl0gJiAyID09IDBACECQUAWgQ29udHJvbCBQYWQgTGlnaHQgU2Vuc29yIChD'
    b'aCA3NilAJ+pQAYMB/0xQFb5FbmFibGUvRGlzYWJsZSBsaWdodCBzZW5zb3IgYXQgNCBIei'
    b'AtIE5PVEU6IERJU0FCTEVEIEZPUiBBTFBIQUEHoFAFnUVuYWJsZSBMaWdodCBTZW5zb3Ig'
    b'KENoIDc2LjApQA+AQE+ASgiASiiASkiA'
)



"""
Default ConfigUI for LOG-0003-200G data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dSSkAoRYFQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABfpQBZNSZW'
    b'NvcmRpbmcgRGlyZWN0b3J5UAGDFP9/UQSGUkVDT1JEUSSBCFAUk0NvbmZpZ1sweDE2RkY3'
    b'Rl09PTBQFbhEaXJlY3RvcnkgZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5ncy4gOCBjaGFyYW'
    b'N0ZXIgbWF4aW11bUAFQKJQBZVSZWNvcmRpbmcgRmlsZSBQcmVmaXhQAYMV/39RJIEIUBXn'
    b'UHJlZml4IGZvciBzZXJpYWxpemVkIHJlY29yZGluZyBuYW1lcy4gSWYgYmxhbmssIGRldm'
    b'ljZSBzZXJpYWwgbnVtYmVyIHdpbGwgYmUgdXNlZC4gOCBjaGFyYWN0ZXIgbWF4aW11bVAU'
    b'k0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGBAF'
    b'AVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIHRo'
    b'ZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1pZG'
    b'5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZWN0'
    b'b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51bW'
    b'JlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBYpU'
    b'aW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQJlQBYtCdXR0b24gTW9kZVABgxD/f1EBgQ'
    b'BQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3Ig'
    b'YmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBTZW'
    b'NvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BAF0EUUAWOUGx1Zy1JbiBBY3Rpb25QAYMK/39R'
    b'AYEBUBWtU2V0IGFjdGlvbiBmb3IgZGV2aWNlIHdoZW4gcGx1Z2dlZCBpbiB0byBVU0IuQQ'
    b'e1UAWySW1tZWRpYXRlbHkgc3RvcCByZWNvcmRpbmcgYW5kIGFwcGVhciBhcyBVU0IgZHJp'
    b'dmVBB7NQBbBDb21wbGV0ZSByZWNvcmRpbmcgYmVmb3JlIGFwcGVhcmluZyBhcyBVU0IgZH'
    b'JpdmVBB7BQBa1JZ25vcmU6IHN0b3AgcmVjb3JkaW5nIHdoZW4gYnV0dG9uIGlzIHByZXNz'
    b'ZWRBB6VQBaJTdGFydCByZWNvcmRpbmcgd2hlbiBVU0IgY29ubmVjdGVkQCPnUAGDC/9/UB'
    b'XeRW50ZXIgbG9jYWwgdGltZXpvbmUncyBvZmZzZXQgZnJvbSBVVEMgdGltZS4gSW1wYWN0'
    b'cyB0aW1lc3RhbXAgaW5mb3JtYXRpb24gb24gcmVjb3JkaW5nIGZpbGVzLkAvgEAPgEBPgE'
    b'AoR25QBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQAYMS/39RAYEAUBVAj0RlbGF5'
    b'IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZSwgdGhlbiB3YWl0IGZvciB0cm'
    b'lnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0ZXIgd2FpdCBvciBvbiB0cmln'
    b'Z2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQeVUAWSRGVsYXkgVGhlbiBUcm'
    b'lnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0YXJ0IGF0IFRpbWVQAYMP/39Q'
    b'FdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdWF0aW5nIG90aGVyIHRyaWdnZX'
    b'IgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5nIGRlbGF5QBGuUAWUUmVjb3Jk'
    b'aW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb25kc0ARQJBQBZhSZWNvcmRpbm'
    b'cgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25kc1EhgwFRgFAV2FRpbWUgYmVm'
    b'b3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZSBiZXR3ZWVuIHJlY29yZGluZ3'
    b'MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlSZWNvcmRpbmcgRmlsZSBTaXpl'
    b'IExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZXRyaWdnZXJQAYMO/39QFLBDb2'
    b'5maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDExRkY3Rl09PW51bGxQFc9XaGVu'
    b'IHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlciByZWNvcmRpbmcgdGltZSBvci'
    b'BzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQgZm9yIEFsbCBTZW5zb3IgQ29u'
    b'ZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbmcgd2lsbCBvbmx5IHN0YXJ0IH'
    b'doZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lzZSByZWNvcmRpbmcgd2lsbCBz'
    b'dGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG1ldEAYQQpQBZBQcmVzc3VyZS'
    b'BUcmlnZ2VyUAGDBQAkUQCBAFAUmUNvbmZpZ1sweDAxRkYyNF0gJiAxID09IDBAguVQBZVQ'
    b'cmVzc3VyZSBUcmlnZ2VyLCBMb3dQAYMDACRRAYMBX5BQJYJQYVEhgwHUwFAVs1NldCB0by'
    b'BzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cmUgd2luZG93LkCC5lAFllBy'
    b'ZXNzdXJlIFRyaWdnZXIsIEhpZ2hQAYMEACRRAYMBrbBQJYJQYVEhgwHUwFAVs1NldCB0by'
    b'BzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cmUgd2luZG93LkAYQRdQBZNU'
    b'ZW1wZXJhdHVyZSBUcmlnZ2VyUAGDBQEkUQCBAFAUmUNvbmZpZ1sweDAxRkYyNF0gJiAyID'
    b'09IDBAY+pQBZhUZW1wZXJhdHVyZSBUcmlnZ2VyLCBMb3dQAYMDASRRAoH2UCWBQ1ESgdhR'
    b'IoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZSB0ZW1wZXJhdHVyZS'
    b'B3aW5kb3cuQGPrUAWZVGVtcGVyYXR1cmUgVHJpZ2dlciwgSGlnaFABgwQBJFECgSNQJYFD'
    b'URKB2FEigVBQFbZTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaWRlIHRlbXBlcm'
    b'F0dXJlIHdpbmRvdy5AGEGdUAWbQWNjZWxlcmF0aW9uIFRyaWdnZXIgKDIwMGcpUAGDB/8g'
    b'UBXUTG93IHBvd2VyIHRyaWdnZXIgdXNpbmcgZGlnaXRhbCBhY2NlbGVyb21ldGVyIHdpdG'
    b'ggREMgcmVzcG9uc2UuIE5vIHByZS10cmlnZ2VyIGRhdGEuUQCBAFARgwX/CFAkgFAUlENv'
    b'bmZpZ1sweDFmZjIwXSA9PSAwQENAsVAFlkFjY2VsZXJhdGlvbiBUaHJlc2hvbGRQAYME/y'
    b'BRAYFQUgOIP+j1wo9cKPZRI4hAaQAAAAAAAFETgFAV8kxvdyBwb3dlciB0cmlnZ2VyLiBU'
    b'aGUgbWluaW11bSBhY2NlbGVyYXRpb24gKHBvc2l0aXZlIG9yIG5lZ2F0aXZlKSB0byB0cm'
    b'lnZ2VyIGEgcmVjb3JkaW5nLiBNdXN0IGJlIGdyZWF0ZXIgdGhhbiAwLkAnxlABgwX/IFEB'
    b'gQdBB5FQBY5YIEF4aXMgVHJpZ2dlckEHkVAFjlkgQXhpcyBUcmlnZ2VyQQeRUAWOWiBBeG'
    b'lzIFRyaWdnZXJAD4BAT4BAKEVGUAWMTWVhc3VyZW1lbnRzQAhAu1AFnDIwMGcgREMgQWNj'
    b'ZWxlcmF0aW9uIChDaCAzMilAJ59QAYMB/yBRAYEBQQeSUAWPRW5hYmxlIEFsbCBBeGVzQB'
    b'f3UAWLU2FtcGxlIFJhdGVQAYMC/yBRAYIBkFAlgkh6UBSTbm90IENvbmZpZ1sweDFmZjIw'
    b'XUEHhVEBggyAQQeFUQGCBkBBB4VRAYIDIEEHhVEBggGQQQeEUQGByEEHhFEBgWRBB4RRAY'
    b'EyQQeEUQGBGUEHhFEBgQxACEEwUAWoQ29udHJvbCBQYWQgVGVtcGVyYXR1cmUvUHJlc3N1'
    b'cmUgKENoIDU5KUAnQJtRAYEDUAGDAf87UBWlRW5hYmxlL0Rpc2FibGUgc2Vuc29ycyBvbi'
    b'Bjb250cm9sIHBhZEEHnFAFmUVuYWJsZSBQcmVzc3VyZSAoQ2ggNTkuMClBB59QBZxFbmFi'
    b'bGUgVGVtcGVyYXR1cmUgKENoIDU5LjEpQQelUAWiRW5hYmxlIFJlbGF0aXZlIEh1bWlkaX'
    b'R5IChDaCA1OS4yKUAR41AFi1NhbXBsZSBSYXRlUAGDAv87UBSUQ29uZmlnWzB4MWZmM0Jd'
    b'ID09IDBQFaRTZXQgc2FtcGxlIGZyZXF1ZW5jeSBmcm9tIDEgdG8gMTAgSHpRAYEKUCWCSH'
    b'pREYEBUSGBCkAIQclQBZlJbmVydGlhbCBNZWFzdXJlbWVudCBVbml0QAdAklAFkEFjcXVp'
    b'c2l0aW9uIE1vZGVQAYMB/ytRAYEQQQeKUAWDT2ZmUQGBAEEHqFAFoUFic29sdXRlIE9yaW'
    b'VudGF0aW9uIChRdWF0ZXJuaW9uKVEBgQhBB6hQBaFSZWxhdGl2ZSBPcmllbnRhdGlvbiAo'
    b'UXVhdGVybmlvbilRAYEQQQePUAWIUm90YXRpb25RAYECQAFAhlAFlU9yaWVudGF0aW9uIE'
    b'RhdGEgUmF0ZVABgwIIK1EBgWRQJYJIelERgQFRIYHIUBW9U2FtcGxlIHJhdGUgZm9yIGRp'
    b'cmVjdGlvbmFsIGRhdGEgZnJvbSAxIHRvIDIwMCBIei4gQ2hhbm5lbCA2NVAUlENvbmZpZ1'
    b'sweDAxRkYyQl0gPCA4QAFAiVAFklJvdGF0aW9uIERhdGEgUmF0ZVABgwICK1EBgWRQJYJI'
    b'elERgQFRIYHIUBW+U2V0IHNhbXBsZSByYXRlIGZvciByb3RhdGlvbiBkYXRhIGZyb20gMS'
    b'B0byAyMDAgSHouIENoYW5uZWwgNDdQFJlDb25maWdbMHgwMUZGMkJdICYgMiA9PSAwQAhA'
    b'8VAFpUludGVybmFsIFRlbXBlcmF0dXJlL1ByZXNzdXJlIChDaCAzNilAJ0DFUQGBA1ABgw'
    b'H/JFAV90VuYWJsZS9EaXNhYmxlIHNsb3dlciBpbnRlcm5hbCBlbnZpcm9ubWVudGFsIHNl'
    b'bnNvcnMuIE5PVEU6IFRlbXBlcmF0dXJlIGlzIHJlcXVpcmVkIGZvciBtYWluIGFjY2VsZX'
    b'JvbWV0ZXIgbWVhc3VyZW1lbnRzQQecUAWZRW5hYmxlIFByZXNzdXJlIChDaCAzNi4wKUEH'
    b'n1AFnEVuYWJsZSBUZW1wZXJhdHVyZSAoQ2ggMzYuMSlACPlQBaBDb250cm9sIFBhZCBMaW'
    b'dodCBTZW5zb3IgKENoIDc2KUAn01EBgQFQAYMB/0xQFaNFbmFibGUvRGlzYWJsZSBsaWdo'
    b'dCBzZW5zb3IgYXQgNCBIekEHoFAFnUVuYWJsZSBMaWdodCBTZW5zb3IgKENoIDc2LjApQA'
    b'+AQE+ASgiASiiASkiA'
)



"""
Linux-specific functions; primarily filesystem-related. The parent package
imports the appropriate version for the host OS.
"""

__author__ = "Connor"

__all__ = ('deviceChanged', 'getDeviceList', 'getBlockSize', 'getFreeSpace',
           'getDriveInfo', 'readRecorderClock', 'readUncachedFile')

import errno
import logging
import os
import mmap
import math
import re
from time import time
from typing import Tuple, Union
import warnings

import psutil

from .types import Drive, Epoch, Filename

logger = logging.getLogger('endaq.device')

# ==============================================================================
#
# ==============================================================================


def _getDeviceIds() -> dict:
    idRoot = '/dev/disk/by-id/'
    return {os.path.abspath(idRoot + os.readlink(idRoot + device)): device
            for device in os.listdir(idRoot)}


def _getDeviceLabels() -> dict:
    idRoot = '/dev/disk/by-label/'
    return {os.path.abspath(idRoot + os.readlink(idRoot + device)): device
            for device in os.listdir(idRoot)}


def getDriveInfo(dev: Filename) -> Union[Drive, None]:
    """ Get general device information. Not currently used.
    """
    dev = os.path.realpath(dev)

    try:
        disk = [x for x in psutil.disk_partitions()
                if x.mountpoint == dev
                if os.path.split(x.device)[-1] in os.listdir('/dev')][0]
        partName = disk.device
    except IndexError:
        return None

    ids = _getDeviceIds()
    if partName not in ids:
        serial = False
    else:
        serialMatch = re.match(r'(.*MIDE.*)_(\d+)-(.*)', ids[partName])
        if serialMatch is None:
            serial = None
        else:
            serial = serialMatch[2]

    labels = _getDeviceLabels()
    if partName not in labels:
        label = None
    else:
        label = labels[partName]

    return Drive(path=dev, label=label, sn=serial, fs=disk.fstype, type=None)


def readUncachedFile(filename: Filename) -> bytes:
    """ Read a file, circumventing the disk cache. Returns the data read.
    """
    filename = os.path.realpath(filename)
    root = os.path.dirname(filename)

    if not os.path.isfile(filename):
        raise IOError(errno.ENOENT, 'No such file', filename)

    # For efficiency, this reads entire blocks.
    # Get the file size, rounded up to bytes per filesystem block.
    size = os.path.getsize(filename)
    blockSize = getBlockSize(root)
    readSize = blockSize*math.ceil(size/blockSize)
    m = mmap.mmap(-1, readSize)

    f = os.open(filename, os.O_RDWR | os.O_DIRECT)

    os.lseek(f, 0, os.SEEK_SET)
    os.readv(f, [m])
    m.seek(0)

    return m.read()


def readRecorderClock(clockFile: Filename,
                      pause: bool = True,
                      timeout: Union[int, float] = 3) -> Tuple[Epoch, Epoch]:
    """ Read a (recorder) clock file, circumventing the disk cache. Returns
        the system time and the encoded device time.

        :param clockFile: The full path to the device's clock file (e.g.,
            `/SYSTEM/DEV/CLOCK`).
        :param pause: If `True` (default), wait until the data in the clock
            file changes (e.g. a new tick/second has started) before
            returning. This is a means to maximize accuracy.
        :param timeout: Seconds to wait for a successful read, when `pause`
            is `True`, before raising a `TimeoutError`.
        :return: The host time, and the unparsed contents of the device clock
            file.
    """
    root = os.path.abspath(os.path.join(os.path.dirname(clockFile), '..', '..'))

    blockSize = getBlockSize(root)
    m = mmap.mmap(-1, blockSize)

    f = os.open(clockFile, os.O_RDWR | os.O_DIRECT)

    os.lseek(f, 0, os.SEEK_SET)
    os.readv(f, [m])
    lastTime = m.read()

    thisTime = lastTime
    sysTime = time()
    deadline = sysTime + timeout
    os.lseek(f, 0, os.SEEK_SET)
    m.seek(0)

    if pause:
        while lastTime == thisTime:
            sysTime = time()
            if timeout > 0 and sysTime > deadline:
                raise TimeoutError('Timed out waiting for CLOCK file to update')
            os.readv(f, [m])
            thisTime = m.read()
            sysTime = (time() + sysTime)/2
            os.lseek(f, 0, os.SEEK_SET)
            m.seek(0)

    os.close(f)

    return sysTime, thisTime


# ==============================================================================
#
# ==============================================================================


def getDeviceList(types: dict, strict: bool = True) -> list:
    """ Get a list of data recorders, as their respective mount points.
    """

    result = set()

    for device, mountpoint, fstype, opts, maxfile, maxpath in psutil.disk_partitions():
        try:
            if not os.path.exists(device):
                continue
            for t in types:
                if t.isRecorder(mountpoint, strict=strict):
                    result.add(mountpoint)
        except IOError as err:
            # Rare error, may be caused by flaky device or USB.
            msg = ("getDeviceList(): Could not access {} ({}); "
                   "ignoring error and continuing".format(device, err))
            warnings.warn(msg)
            logger.error(msg)

    return sorted(result)


# Module-level globals for caching last discovered logical drives and recorders
_LAST_DEVICES = None  # List of sdiskpart namedtuples from psutils
_LAST_RECORDERS = None  # tuple of mountpoints of last seen endaq devices


def deviceChanged(recordersOnly: bool, types: dict, clear: bool = False) -> bool:
    """ Returns `True` if a drive has been connected or disconnected since
        the last call to `deviceChanged()`.

        :param recordersOnly: If `False`, any change to the mounted drives
            is reported as a change. If `True`, the mounted drives are checked
            and `True` is only returned if the change occurred to a recorder.
            Checking for recorders only takes marginally more time.
        :param types: A list of known `Recorder` classes to detect.
        :param clear: If `True`, clear the cache of previously-detected
            drives and devices.
    """
    global _LAST_DEVICES, _LAST_RECORDERS

    if clear:
        _LAST_DEVICES = 0
        _LAST_RECORDERS = None

    newDevices = psutil.disk_partitions()
    changed = newDevices != _LAST_DEVICES
    _LAST_DEVICES = newDevices

    # if not changed or not recordersOnly:
    if not recordersOnly:
        return changed

    newRecorders = tuple(getDeviceList(types=types))
    changed = newRecorders != _LAST_RECORDERS
    _LAST_RECORDERS = newRecorders
    return changed


# ==============================================================================
#
# ==============================================================================


def getFreeSpace(path: Filename) -> int:
    """ Return the free space (in bytes) on a drive.

        :param path: The path to the drive to check. Can be a subdirectory.
        :return: The free space on the drive, in bytes.
        :rtype: int
    """
    return psutil.disk_usage(path).free


def getBlockSize(path: Filename) -> int:
    """ Return the bytes per sector and sectors per cluster of a drive.

        :param path: The path to the drive to check. Can be a subdirectory.
        :return: A tuple containing the bytes/sector and sectors/cluster.
    """
    return os.statvfs(path).f_bsize



"""
Default ConfigUI for S3-D16 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dSSEAoRYFQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABfpQBZNSZW'
    b'NvcmRpbmcgRGlyZWN0b3J5UAGDFP9/UQSGUkVDT1JEUSSBCFAUk0NvbmZpZ1sweDE2RkY3'
    b'Rl09PTBQFbhEaXJlY3RvcnkgZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5ncy4gOCBjaGFyYW'
    b'N0ZXIgbWF4aW11bUAFQKJQBZVSZWNvcmRpbmcgRmlsZSBQcmVmaXhQAYMV/39RJIEIUBXn'
    b'UHJlZml4IGZvciBzZXJpYWxpemVkIHJlY29yZGluZyBuYW1lcy4gSWYgYmxhbmssIGRldm'
    b'ljZSBzZXJpYWwgbnVtYmVyIHdpbGwgYmUgdXNlZC4gOCBjaGFyYWN0ZXIgbWF4aW11bVAU'
    b'k0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGBAF'
    b'AVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIHRo'
    b'ZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1pZG'
    b'5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZWN0'
    b'b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51bW'
    b'JlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBYpU'
    b'aW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQJlQBYtCdXR0b24gTW9kZVABgxD/f1EBgQ'
    b'BQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3Ig'
    b'YmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBTZW'
    b'NvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BAF0EUUAWOUGx1Zy1JbiBBY3Rpb25QAYMK/39R'
    b'AYEBUBWtU2V0IGFjdGlvbiBmb3IgZGV2aWNlIHdoZW4gcGx1Z2dlZCBpbiB0byBVU0IuQQ'
    b'e1UAWySW1tZWRpYXRlbHkgc3RvcCByZWNvcmRpbmcgYW5kIGFwcGVhciBhcyBVU0IgZHJp'
    b'dmVBB7NQBbBDb21wbGV0ZSByZWNvcmRpbmcgYmVmb3JlIGFwcGVhcmluZyBhcyBVU0IgZH'
    b'JpdmVBB7BQBa1JZ25vcmU6IHN0b3AgcmVjb3JkaW5nIHdoZW4gYnV0dG9uIGlzIHByZXNz'
    b'ZWRBB6VQBaJTdGFydCByZWNvcmRpbmcgd2hlbiBVU0IgY29ubmVjdGVkQCPnUAGDC/9/UB'
    b'XeRW50ZXIgbG9jYWwgdGltZXpvbmUncyBvZmZzZXQgZnJvbSBVVEMgdGltZS4gSW1wYWN0'
    b'cyB0aW1lc3RhbXAgaW5mb3JtYXRpb24gb24gcmVjb3JkaW5nIGZpbGVzLkAvgEAPgEBPgE'
    b'AoR21QBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQAYMS/39RAYEAUBVAj0RlbGF5'
    b'IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZSwgdGhlbiB3YWl0IGZvciB0cm'
    b'lnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0ZXIgd2FpdCBvciBvbiB0cmln'
    b'Z2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQeVUAWSRGVsYXkgVGhlbiBUcm'
    b'lnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0YXJ0IGF0IFRpbWVQAYMP/39Q'
    b'FdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdWF0aW5nIG90aGVyIHRyaWdnZX'
    b'IgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5nIGRlbGF5QBGuUAWUUmVjb3Jk'
    b'aW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb25kc0ARQJBQBZhSZWNvcmRpbm'
    b'cgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25kc1EhgwFRgFAV2FRpbWUgYmVm'
    b'b3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZSBiZXR3ZWVuIHJlY29yZGluZ3'
    b'MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlSZWNvcmRpbmcgRmlsZSBTaXpl'
    b'IExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZXRyaWdnZXJQAYMO/39QFLBDb2'
    b'5maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDExRkY3Rl09PW51bGxQFc9XaGVu'
    b'IHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlciByZWNvcmRpbmcgdGltZSBvci'
    b'BzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQgZm9yIEFsbCBTZW5zb3IgQ29u'
    b'ZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbmcgd2lsbCBvbmx5IHN0YXJ0IH'
    b'doZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lzZSByZWNvcmRpbmcgd2lsbCBz'
    b'dGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG1ldEAYQQpQBZBQcmVzc3VyZS'
    b'BUcmlnZ2VyUAGDBQAkUQCBAFAUmUNvbmZpZ1sweDAxRkYyNF0gJiAxID09IDBAguVQBZVQ'
    b'cmVzc3VyZSBUcmlnZ2VyLCBMb3dQAYMDACRRAYMBX5BQJYJQYVEhgwHUwFAVs1NldCB0by'
    b'BzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cmUgd2luZG93LkCC5lAFllBy'
    b'ZXNzdXJlIFRyaWdnZXIsIEhpZ2hQAYMEACRRAYMBrbBQJYJQYVEhgwHUwFAVs1NldCB0by'
    b'BzdGFydCBzYW1wbGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cmUgd2luZG93LkAYQRdQBZNU'
    b'ZW1wZXJhdHVyZSBUcmlnZ2VyUAGDBQEkUQCBAFAUmUNvbmZpZ1sweDAxRkYyNF0gJiAyID'
    b'09IDBAY+pQBZhUZW1wZXJhdHVyZSBUcmlnZ2VyLCBMb3dQAYMDASRRAoH2UCWBQ1ESgdhR'
    b'IoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4gb3V0c2lkZSB0ZW1wZXJhdHVyZS'
    b'B3aW5kb3cuQGPrUAWZVGVtcGVyYXR1cmUgVHJpZ2dlciwgSGlnaFABgwQBJFECgSNQJYFD'
    b'URKB2FEigVBQFbZTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaWRlIHRlbXBlcm'
    b'F0dXJlIHdpbmRvdy5AGEGcUAWaQWNjZWxlcmF0aW9uIFRyaWdnZXIgKDE2ZylQAYMH/yBQ'
    b'FdRMb3cgcG93ZXIgdHJpZ2dlciB1c2luZyBkaWdpdGFsIGFjY2VsZXJvbWV0ZXIgd2l0aC'
    b'BEQyByZXNwb25zZS4gTm8gcHJlLXRyaWdnZXIgZGF0YS5RAIEAUBGDBf8IUCSAUBSUQ29u'
    b'ZmlnWzB4MWZmMjBdID09IDBAQ0CxUAWWQWNjZWxlcmF0aW9uIFRocmVzaG9sZFABgwT/IF'
    b'EBgVBSA4g/sBAQEBAQEFEjiEAwAAAAAAAAUROAUBXyTG93IHBvd2VyIHRyaWdnZXIuIFRo'
    b'ZSBtaW5pbXVtIGFjY2VsZXJhdGlvbiAocG9zaXRpdmUgb3IgbmVnYXRpdmUpIHRvIHRyaW'
    b'dnZXIgYSByZWNvcmRpbmcuIE11c3QgYmUgZ3JlYXRlciB0aGFuIDAuQCfGUAGDBf8gUQGB'
    b'B0EHkVAFjlggQXhpcyBUcmlnZ2VyQQeRUAWOWSBBeGlzIFRyaWdnZXJBB5FQBY5aIEF4aX'
    b'MgVHJpZ2dlckAPgEBPgEAoRUVQBYxNZWFzdXJlbWVudHNACEC6UAWbMTZnIERDIEFjY2Vs'
    b'ZXJhdGlvbiAoQ2ggMzIpQCefUAGDAf8gUQGBAUEHklAFj0VuYWJsZSBBbGwgQXhlc0AX91'
    b'AFi1NhbXBsZSBSYXRlUAGDAv8gUQGCAZBQJYJIelAUk25vdCBDb25maWdbMHgxZmYyMF1B'
    b'B4VRAYIMgEEHhVEBggZAQQeFUQGCAyBBB4VRAYIBkEEHhFEBgchBB4RRAYFkQQeEUQGBMk'
    b'EHhFEBgRlBB4RRAYEMQAhBMFAFqENvbnRyb2wgUGFkIFRlbXBlcmF0dXJlL1ByZXNzdXJl'
    b'IChDaCA1OSlAJ0CbUQGBA1ABgwH/O1AVpUVuYWJsZS9EaXNhYmxlIHNlbnNvcnMgb24gY2'
    b'9udHJvbCBwYWRBB5xQBZlFbmFibGUgUHJlc3N1cmUgKENoIDU5LjApQQefUAWcRW5hYmxl'
    b'IFRlbXBlcmF0dXJlIChDaCA1OS4xKUEHpVAFokVuYWJsZSBSZWxhdGl2ZSBIdW1pZGl0eS'
    b'AoQ2ggNTkuMilAEeNQBYtTYW1wbGUgUmF0ZVABgwL/O1AUlENvbmZpZ1sweDFmZjNCXSA9'
    b'PSAwUBWkU2V0IHNhbXBsZSBmcmVxdWVuY3kgZnJvbSAxIHRvIDEwIEh6UQGBClAlgkh6UR'
    b'GBAVEhgQpACEHJUAWZSW5lcnRpYWwgTWVhc3VyZW1lbnQgVW5pdEAHQJJQBZBBY3F1aXNp'
    b'dGlvbiBNb2RlUAGDAf8rUQGBEEEHilAFg09mZlEBgQBBB6hQBaFBYnNvbHV0ZSBPcmllbn'
    b'RhdGlvbiAoUXVhdGVybmlvbilRAYEIQQeoUAWhUmVsYXRpdmUgT3JpZW50YXRpb24gKFF1'
    b'YXRlcm5pb24pUQGBEEEHj1AFiFJvdGF0aW9uUQGBAkABQIZQBZVPcmllbnRhdGlvbiBEYX'
    b'RhIFJhdGVQAYMCCCtRAYFkUCWCSHpREYEBUSGByFAVvVNhbXBsZSByYXRlIGZvciBkaXJl'
    b'Y3Rpb25hbCBkYXRhIGZyb20gMSB0byAyMDAgSHouIENoYW5uZWwgNjVQFJRDb25maWdbMH'
    b'gwMUZGMkJdIDwgOEABQIlQBZJSb3RhdGlvbiBEYXRhIFJhdGVQAYMCAitRAYFkUCWCSHpR'
    b'EYEBUSGByFAVvlNldCBzYW1wbGUgcmF0ZSBmb3Igcm90YXRpb24gZGF0YSBmcm9tIDEgdG'
    b'8gMjAwIEh6LiBDaGFubmVsIDQ3UBSZQ29uZmlnWzB4MDFGRjJCXSAmIDIgPT0gMEAIQPFQ'
    b'BaVJbnRlcm5hbCBUZW1wZXJhdHVyZS9QcmVzc3VyZSAoQ2ggMzYpQCdAxVEBgQNQAYMB/y'
    b'RQFfdFbmFibGUvRGlzYWJsZSBzbG93ZXIgaW50ZXJuYWwgZW52aXJvbm1lbnRhbCBzZW5z'
    b'b3JzLiBOT1RFOiBUZW1wZXJhdHVyZSBpcyByZXF1aXJlZCBmb3IgbWFpbiBhY2NlbGVyb2'
    b'1ldGVyIG1lYXN1cmVtZW50c0EHnFAFmUVuYWJsZSBQcmVzc3VyZSAoQ2ggMzYuMClBB59Q'
    b'BZxFbmFibGUgVGVtcGVyYXR1cmUgKENoIDM2LjEpQAj5UAWgQ29udHJvbCBQYWQgTGlnaH'
    b'QgU2Vuc29yIChDaCA3NilAJ9NRAYEBUAGDAf9MUBWjRW5hYmxlL0Rpc2FibGUgbGlnaHQg'
    b'c2Vuc29yIGF0IDQgSHpBB6BQBZ1FbmFibGUgTGlnaHQgU2Vuc29yIChDaCA3Ni4wKUAPgE'
    b'BPgEoIgEoogEpIgA=='
)



"""
Default ConfigUI for S5-E2000D40 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dViUAoRZNQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABftQBZNSZW'
    b'NvcmRpbmcgRGlyZWN0b3J5UAGDFP9/UQSGUkVDT1JEUSSBEFAUk0NvbmZpZ1sweDE2RkY3'
    b'Rl09PTBQFblEaXJlY3RvcnkgZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5ncy4gMTYgY2hhcm'
    b'FjdGVyIG1heGltdW1ABUCjUAWVUmVjb3JkaW5nIEZpbGUgUHJlZml4UAGDFf9/USSBEFAV'
    b'6FByZWZpeCBmb3Igc2VyaWFsaXplZCByZWNvcmRpbmcgbmFtZXMuIElmIGJsYW5rLCBkZX'
    b'ZpY2Ugc2VyaWFsIG51bWJlciB3aWxsIGJlIHVzZWQuIDE2IGNoYXJhY3RlciBtYXhpbXVt'
    b'UBSTQ29uZmlnWzB4MTZGRjdGXT09MEAHQTxQBY9GaWxlIE5hbWUgU3R5bGVQAYMW/39RAY'
    b'EBUBVA/FRpbWUtQmFzZWQgZmlsZSBuYW1lcyB1c2UgYSBkaXJlY3RvcnkgYmFzZWQgb24g'
    b'dGhlIGRhdGUsIGFuZCBmaWxlIG5hbWUgYmFzZWQgb24gdGhlIHNlY29uZHMgc2luY2UgbW'
    b'lkbmlnaHQuClNlcmlhbGl6ZWQgZmlsZSBuYW1lcyB1c2UgdGhlIHNwZWNpZmllZCBkaXJl'
    b'Y3RvcnkgYW5kIGZpbGUgbmFtZSBwcmVmaXgsIGZvbGxvd2VkIGJ5IGEgbnVtYmVyLiAKTn'
    b'VtYmVyIGlzIHJlc2V0IHdoZW4gZGlyZWN0b3J5IG9yIHByZWZpeCBjaGFuZ2VzLkEHjVAF'
    b'ilRpbWUgQmFzZWRBB41QBYpTZXJpYWxpemVkQAdAqVAFi0J1dHRvbiBNb2RlUAGDEP9/UQ'
    b'GBAFAVv09uZSBTZWNvbmQgUHJlc3Mgd2lsbCBkaXNhYmxlIDMtc2Vjb25kIHByZXNzIGZv'
    b'ciBiYXR0ZXJ5IGNoZWNrLkEHl1AFlEltbWVkaWF0ZSBTdGFydC9TdG9wQQeiUAWfT25lIF'
    b'NlY29uZCBQcmVzcyBmb3IgU3RhcnQvU3RvcEEHjVAFilN0YXJ0LU9ubHlAF0EUUAWOUGx1'
    b'Zy1JbiBBY3Rpb25QAYMK/39RAYEBUBWtU2V0IGFjdGlvbiBmb3IgZGV2aWNlIHdoZW4gcG'
    b'x1Z2dlZCBpbiB0byBVU0IuQQe1UAWySW1tZWRpYXRlbHkgc3RvcCByZWNvcmRpbmcgYW5k'
    b'IGFwcGVhciBhcyBVU0IgZHJpdmVBB7NQBbBDb21wbGV0ZSByZWNvcmRpbmcgYmVmb3JlIG'
    b'FwcGVhcmluZyBhcyBVU0IgZHJpdmVBB7BQBa1JZ25vcmU6IHN0b3AgcmVjb3JkaW5nIHdo'
    b'ZW4gYnV0dG9uIGlzIHByZXNzZWRBB6VQBaJTdGFydCByZWNvcmRpbmcgd2hlbiBVU0IgY2'
    b'9ubmVjdGVkQCPnUAGDC/9/UBXeRW50ZXIgbG9jYWwgdGltZXpvbmUncyBvZmZzZXQgZnJv'
    b'bSBVVEMgdGltZS4gSW1wYWN0cyB0aW1lc3RhbXAgaW5mb3JtYXRpb24gb24gcmVjb3JkaW'
    b'5nIGZpbGVzLkAvgEAPgEBPgEAoSOlQBYhUcmlnZ2Vyc0AHQNpQBYxUcmlnZ2VyIE1vZGVQ'
    b'AYMS/39RAYEAUBVAj0RlbGF5IFRoZW4gVHJpZ2dlcjogV2FpdCBzcGVjaWZpZWQgdGltZS'
    b'wgdGhlbiB3YWl0IGZvciB0cmlnZ2VyCkRlbGF5IE9yIFRyaWdnZXI6IEFjcXVpcmUgYWZ0'
    b'ZXIgd2FpdCBvciBvbiB0cmlnZ2VyLiBCYXR0ZXJ5IGxpZmUgbWF5IGJlIHJlZHVjZWQuQQ'
    b'eVUAWSRGVsYXkgVGhlbiBUcmlnZ2VyQQeTUAWQRGVsYXkgT3IgVHJpZ2dlckAy9FAFjVN0'
    b'YXJ0IGF0IFRpbWVQAYMP/39QFdtXYWl0IHVudGlsIHNldCB0aW1lIGJlZm9yZSBldmFsdW'
    b'F0aW5nIG90aGVyIHRyaWdnZXIgY29uZGl0aW9ucywgaW5jbHVkaW5nIGEgcmVjb3JkaW5n'
    b'IGRlbGF5QBGuUAWUUmVjb3JkaW5nIFRpbWUgTGltaXRQAYMN/39RIYT////wUCWHU2Vjb2'
    b'5kc0ARQJBQBZhSZWNvcmRpbmcgRGVsYXkvSW50ZXJ2YWxQAYMM/39RAYEAUCWHU2Vjb25k'
    b'c1EhgwFRgFAV2FRpbWUgYmVmb3JlIHN0YXJ0aW5nIGEgcmVjb3JkaW5nLCBhbmQgdGltZS'
    b'BiZXR3ZWVuIHJlY29yZGluZ3MgaWYgJ1JldHJpZ2dlcicgaXMgY2hlY2tlZC5AEa9QBZlS'
    b'ZWNvcmRpbmcgRmlsZSBTaXplIExpbWl0UAGDEf9/USGE////8FAlg0tpQkAQQJdQBYlSZX'
    b'RyaWdnZXJQAYMO/39QFLBDb25maWdbMHhERkY3Rl09PW51bGwgYW5kIENvbmZpZ1sweDEx'
    b'RkY3Rl09PW51bGxQFc9XaGVuIHNldCwgU2xhbSBTdGljayB3aWxsIHJlLWFybSBhZnRlci'
    b'ByZWNvcmRpbmcgdGltZSBvciBzaXplIGxpbWl0IGlzIHJlYWNoZWQuQBBAqFAFnldhaXQg'
    b'Zm9yIEFsbCBTZW5zb3IgQ29uZGl0aW9uc1ABgxP/f1AV/ldoZW4gc2V0LCByZWNvcmRpbm'
    b'cgd2lsbCBvbmx5IHN0YXJ0IHdoZW4gYWxsIHRyaWdnZXJzIGFyZSBtZXQsIG90aGVyd2lz'
    b'ZSByZWNvcmRpbmcgd2lsbCBzdGFydCB3aGVuIGFueSBzZW5zb3IgY29uZGl0aW9uIGlzIG'
    b'1ldEAYQO5QBZBQcmVzc3VyZSBUcmlnZ2VyUAGDBQAkUQCBAECC5VAFlVByZXNzdXJlIFRy'
    b'aWdnZXIsIExvd1ABgwMAJFEBgwFfkFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbX'
    b'BsaW5nIHdoZW4gb3V0c2lkZSBwcmVzc3VyZSB3aW5kb3cuQILmUAWWUHJlc3N1cmUgVHJp'
    b'Z2dlciwgSGlnaFABgwQAJFEBgwGtsFAlglBhUSGDAdTAUBWzU2V0IHRvIHN0YXJ0IHNhbX'
    b'BsaW5nIHdoZW4gb3V0c2lkZSBwcmVzc3VyZSB3aW5kb3cuQBhA+1AFk1RlbXBlcmF0dXJl'
    b'IFRyaWdnZXJQAYMFASRRAIEAQGPqUAWYVGVtcGVyYXR1cmUgVHJpZ2dlciwgTG93UAGDAw'
    b'EkUQKB9lAlgUNREoHYUSKBUFAVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91dHNp'
    b'ZGUgdGVtcGVyYXR1cmUgd2luZG93LkBj61AFmVRlbXBlcmF0dXJlIFRyaWdnZXIsIEhpZ2'
    b'hQAYMEASRRAoEjUCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdoZW4g'
    b'b3V0c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQBhBqVAFnEFjY2VsZXJhdGlvbiBUcmlnZ2'
    b'VyICgyMDAwZylQAYMF/whQEYMH/yBRAIEAUCSAUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAw'
    b'QENAq1AFmUFjY2VsZXJhdGlvbiBUcmlnZ2VyLCBMb3dQAYMD/whRAoL5mlETiMCfQAAAAA'
    b'AAUSOIQJ9AAAAAAABQFeBIaWdoIHBvd2VyIHRyaWdnZXIgd2l0aCBwcmUtdHJpZ2dlciBk'
    b'YXRhLiBOb3QgcmVjb21tZW5kZWQgZm9yIGxvbmcgd2FpdHMgd2l0aG91dCBleHRlcm5hbC'
    b'Bwb3dlci5SA4g/r0AfQB9ILkBDQKxQBZpBY2NlbGVyYXRpb24gVHJpZ2dlciwgSGlnaFAB'
    b'gwT/CFECggZmUROIwJ9AAAAAAABRI4hAn0AAAAAAAFAV4EhpZ2ggcG93ZXIgdHJpZ2dlci'
    b'B3aXRoIHByZS10cmlnZ2VyIGRhdGEuIE5vdCByZWNvbW1lbmRlZCBmb3IgbG9uZyB3YWl0'
    b'cyB3aXRob3V0IGV4dGVybmFsIHBvd2VyLlIDiD+vQB9AH0guQBhBo1AFmkFjY2VsZXJhdG'
    b'lvbiBUcmlnZ2VyICg0MGcpUAGDB/9QUBXUTG93IHBvd2VyIHRyaWdnZXIgdXNpbmcgZGln'
    b'aXRhbCBhY2NlbGVyb21ldGVyIHdpdGggREMgcmVzcG9uc2UuIE5vIHByZS10cmlnZ2VyIG'
    b'RhdGEuUQCBAFARgwX/CFARgwf/IFAkgFAUlENvbmZpZ1sweDFmZjUwXSA9PSAwQENAslAF'
    b'lkFjY2VsZXJhdGlvbiBUaHJlc2hvbGRQAYME/1BRAYIgAFIDiD9EABQAFAAUUSOIQEQAAA'
    b'AAAABRE4BQFfJMb3cgcG93ZXIgdHJpZ2dlci4gVGhlIG1pbmltdW0gYWNjZWxlcmF0aW9u'
    b'IChwb3NpdGl2ZSBvciBuZWdhdGl2ZSkgdG8gdHJpZ2dlciBhIHJlY29yZGluZy4gTXVzdC'
    b'BiZSBncmVhdGVyIHRoYW4gMC5AJ8ZQAYMF/1BRAYEHQQeRUAWOWCBBeGlzIFRyaWdnZXJB'
    b'B5FQBY5ZIEF4aXMgVHJpZ2dlckEHkVAFjlogQXhpcyBUcmlnZ2VyQA+AQE+AQChG+FAFjE'
    b'1lYXN1cmVtZW50c0AIQkJQBZkyMDAwZyBBY2NlbGVyYXRpb24gKENoIDgpUCSAQCdAgVEB'
    b'gQdQAYMB/whQFaBFbmFibGUvRGlzYWJsZSBhY2NlbGVyYXRpb24gYXhlc0EHmVAFlkVuYW'
    b'JsZSBYIEF4aXMgKENoIDguMClBB5lQBZZFbmFibGUgWSBBeGlzIChDaCA4LjEpQQeZUAWW'
    b'RW5hYmxlIFogQXhpcyAoQ2ggOC4yKUARQJpQBYtTYW1wbGUgUmF0ZVABgwL/CFAV2FNldC'
    b'BtYWluIGFjY2VsZXJhdGlvbiBzYW1wbGUgcmF0ZSBmcm9tIDEwIHRvIDIwMDAwIEh6LgpI'
    b'aWdoZXIgcmF0ZSByZWR1Y2VzIGJhdHRlcnkgbGlmZS5QFJVDb25maWdbMHgwMUZGMDhdID'
    b'09IDBRAYITiFAlgkh6URGBClEhgk4gQBFA/FAFo092ZXJyaWRlIEFudGlhbGlhc2luZyBG'
    b'aWx0ZXIgQ3V0b2ZmUAGDCP8IUQGCA+hQJYJIelEhgmGoUBVApURlZmF1bHQsIHdoZW4gbm'
    b'90IGNoZWNrZWQsIG9mIG9uZS1maWZ0aCBzYW1wbGUgZnJlcXVlbmN5LiBDdXQtb2ZmIGZy'
    b'ZXF1ZW5jeSBoYXMgMjklIGF0dGVudWF0aW9uLgo1dGggb3JkZXIgQnV0dGVyd29ydGguIF'
    b'NvbWUgYXR0ZW51YXRpb24gYXQgNjAlIG9mIGN1dC1vZmYgZnJlcXVlbmN5LlAUlUNvbmZp'
    b'Z1sweDAxRkYwOF0gPT0gMEAIQRxQBZs0MGcgREMgQWNjZWxlcmF0aW9uIChDaCA4MClAJ5'
    b'9QAYMB/1BRAYEBQQeSUAWPRW5hYmxlIEFsbCBBeGVzQBdA2FAFi1NhbXBsZSBSYXRlUAGD'
    b'Av9QUQGCAfRQJYJIelAV3kxvdyBwYXNzIGZpbHRlciBpcyBzZXQgdG8gMjUlIG9mIGRhdG'
    b'EgcmF0ZSwgc28gYSBkYXRhIHJhdGUgb2YgMjAwMEh6IGhhcyBhIExQIGZpbHRlciBhdCA1'
    b'MDAgSHpQFJNub3QgQ29uZmlnWzB4MWZmNTBdQQeFUQGCD6BBB4VRAYIH0EEHhVEBggPoQQ'
    b'eFUQGCAfRBB4RRAYH6QQeEUQGBfUEHhFEBgT9BB4RRAYEgQQeEUQGBEEAIQTBQBahDb250'
    b'cm9sIFBhZCBUZW1wZXJhdHVyZS9QcmVzc3VyZSAoQ2ggNTkpQCdAm1EBgQNQAYMB/ztQFa'
    b'VFbmFibGUvRGlzYWJsZSBzZW5zb3JzIG9uIGNvbnRyb2wgcGFkQQecUAWZRW5hYmxlIFBy'
    b'ZXNzdXJlIChDaCA1OS4wKUEHn1AFnEVuYWJsZSBUZW1wZXJhdHVyZSAoQ2ggNTkuMSlBB6'
    b'VQBaJFbmFibGUgUmVsYXRpdmUgSHVtaWRpdHkgKENoIDU5LjIpQBHjUAWLU2FtcGxlIFJh'
    b'dGVQAYMC/ztQFJRDb25maWdbMHgxZmYzQl0gPT0gMFAVpFNldCBzYW1wbGUgZnJlcXVlbm'
    b'N5IGZyb20gMSB0byAxMCBIelEBgQpQJYJIelERgQFRIYEKQAhBzVAFmUluZXJ0aWFsIE1l'
    b'YXN1cmVtZW50IFVuaXRAB0CWUAWQQWNxdWlzaXRpb24gTW9kZVABgwH/K1AlgSBRAYEQQQ'
    b'eKUAWDT2ZmUQGBAEEHqFAFoUFic29sdXRlIE9yaWVudGF0aW9uIChRdWF0ZXJuaW9uKVEB'
    b'gQhBB6hQBaFSZWxhdGl2ZSBPcmllbnRhdGlvbiAoUXVhdGVybmlvbilRAYEQQQePUAWIUm'
    b'90YXRpb25RAYECQAFAhlAFlU9yaWVudGF0aW9uIERhdGEgUmF0ZVABgwIIK1EBgWRQJYJI'
    b'elERgQFRIYHIUBW9U2FtcGxlIHJhdGUgZm9yIGRpcmVjdGlvbmFsIGRhdGEgZnJvbSAxIH'
    b'RvIDIwMCBIei4gQ2hhbm5lbCA2NVAUlENvbmZpZ1sweDAxRkYyQl0gPCA4QAFAiVAFklJv'
    b'dGF0aW9uIERhdGEgUmF0ZVABgwICK1EBgWRQJYJIelERgQFRIYHIUBW+U2V0IHNhbXBsZS'
    b'ByYXRlIGZvciByb3RhdGlvbiBkYXRhIGZyb20gMSB0byAyMDAgSHouIENoYW5uZWwgNDdQ'
    b'FJlDb25maWdbMHgwMUZGMkJdICYgMiA9PSAwQAj1UAWgQ29udHJvbCBQYWQgTGlnaHQgU2'
    b'Vuc29yIChDaCA3NilAJ89QAYMB/0xQFaNFbmFibGUvRGlzYWJsZSBsaWdodCBzZW5zb3Ig'
    b'YXQgNCBIekEHoFAFnUVuYWJsZSBMaWdodCBTZW5zb3IgKENoIDc2LjApQA+AQE+ASgiASi'
    b'iASkiA'
)



"""
Default ConfigUI for S2-D8D16 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dV3kAoRjBQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABUCZUAWVQ3'
    b'VzdG9tIFJlY29yZGluZyBUYWdzUAGDF/9/UQSAUSSBQFAV8UtleXdvcmRzIGZvciBvcmdh'
    b'bml6aW5nIGFuZCBpZGVudGlmeWluZyByZWNvcmRpbmdzLCBzZXBhcmF0ZWQgYnkgY29tbW'
    b'Egb3Igc3BhY2UuIExpbWl0IGlzIDIwIHRhZ3MsIDY0IGNoYXJhY3RlcnMuQAX7UAWTUmVj'
    b'b3JkaW5nIERpcmVjdG9yeVABgxT/f1EEhlJFQ09SRFEkgRBQFJNDb25maWdbMHgxNkZGN0'
    b'ZdPT0wUBW5RGlyZWN0b3J5IGZvciBzZXJpYWxpemVkIHJlY29yZGluZ3MuIDE2IGNoYXJh'
    b'Y3RlciBtYXhpbXVtQAVAo1AFlVJlY29yZGluZyBGaWxlIFByZWZpeFABgxX/f1EkgRBQFe'
    b'hQcmVmaXggZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5nIG5hbWVzLiBJZiBibGFuaywgZGV2'
    b'aWNlIHNlcmlhbCBudW1iZXIgd2lsbCBiZSB1c2VkLiAxNiBjaGFyYWN0ZXIgbWF4aW11bV'
    b'AUk0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGB'
    b'AVAVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIH'
    b'RoZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1p'
    b'ZG5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZW'
    b'N0b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51'
    b'bWJlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBY'
    b'pUaW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQKlQBYtCdXR0b24gTW9kZVABgxD/f1EB'
    b'gQBQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3'
    b'IgYmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBT'
    b'ZWNvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BBB41QBYpTdGFydC1Pbmx5QBdBFFAFjlBsdW'
    b'ctSW4gQWN0aW9uUAGDCv9/UQGBAVAVrVNldCBhY3Rpb24gZm9yIGRldmljZSB3aGVuIHBs'
    b'dWdnZWQgaW4gdG8gVVNCLkEHtVAFskltbWVkaWF0ZWx5IHN0b3AgcmVjb3JkaW5nIGFuZC'
    b'BhcHBlYXIgYXMgVVNCIGRyaXZlQQezUAWwQ29tcGxldGUgcmVjb3JkaW5nIGJlZm9yZSBh'
    b'cHBlYXJpbmcgYXMgVVNCIGRyaXZlQQewUAWtSWdub3JlOiBzdG9wIHJlY29yZGluZyB3aG'
    b'VuIGJ1dHRvbiBpcyBwcmVzc2VkQQelUAWiU3RhcnQgcmVjb3JkaW5nIHdoZW4gVVNCIGNv'
    b'bm5lY3RlZEAj51ABgwv/f1AV3kVudGVyIGxvY2FsIHRpbWV6b25lJ3Mgb2Zmc2V0IGZyb2'
    b'0gVVRDIHRpbWUuIEltcGFjdHMgdGltZXN0YW1wIGluZm9ybWF0aW9uIG9uIHJlY29yZGlu'
    b'ZyBmaWxlcy5AL4BAD4BAT4BAKEkZUAWIVHJpZ2dlcnNAB0DaUAWMVHJpZ2dlciBNb2RlUA'
    b'GDEv9/UQGBAFAVQI9EZWxheSBUaGVuIFRyaWdnZXI6IFdhaXQgc3BlY2lmaWVkIHRpbWUs'
    b'IHRoZW4gd2FpdCBmb3IgdHJpZ2dlcgpEZWxheSBPciBUcmlnZ2VyOiBBY3F1aXJlIGFmdG'
    b'VyIHdhaXQgb3Igb24gdHJpZ2dlci4gQmF0dGVyeSBsaWZlIG1heSBiZSByZWR1Y2VkLkEH'
    b'lVAFkkRlbGF5IFRoZW4gVHJpZ2dlckEHk1AFkERlbGF5IE9yIFRyaWdnZXJAMvRQBY1TdG'
    b'FydCBhdCBUaW1lUAGDD/9/UBXbV2FpdCB1bnRpbCBzZXQgdGltZSBiZWZvcmUgZXZhbHVh'
    b'dGluZyBvdGhlciB0cmlnZ2VyIGNvbmRpdGlvbnMsIGluY2x1ZGluZyBhIHJlY29yZGluZy'
    b'BkZWxheUARrlAFlFJlY29yZGluZyBUaW1lIExpbWl0UAGDDf9/USGE////8FAlh1NlY29u'
    b'ZHNAEUCQUAWYUmVjb3JkaW5nIERlbGF5L0ludGVydmFsUAGDDP9/UQGBAFAlh1NlY29uZH'
    b'NRIYMBUYBQFdhUaW1lIGJlZm9yZSBzdGFydGluZyBhIHJlY29yZGluZywgYW5kIHRpbWUg'
    b'YmV0d2VlbiByZWNvcmRpbmdzIGlmICdSZXRyaWdnZXInIGlzIGNoZWNrZWQuQBGvUAWZUm'
    b'Vjb3JkaW5nIEZpbGUgU2l6ZSBMaW1pdFABgxH/f1EhhP////BQJYNLaUJAEECXUAWJUmV0'
    b'cmlnZ2VyUAGDDv9/UBSwQ29uZmlnWzB4REZGN0ZdPT1udWxsIGFuZCBDb25maWdbMHgxMU'
    b'ZGN0ZdPT1udWxsUBXPV2hlbiBzZXQsIFNsYW0gU3RpY2sgd2lsbCByZS1hcm0gYWZ0ZXIg'
    b'cmVjb3JkaW5nIHRpbWUgb3Igc2l6ZSBsaW1pdCBpcyByZWFjaGVkLkAQQKhQBZ5XYWl0IG'
    b'ZvciBBbGwgU2Vuc29yIENvbmRpdGlvbnNQAYMT/39QFf5XaGVuIHNldCwgcmVjb3JkaW5n'
    b'IHdpbGwgb25seSBzdGFydCB3aGVuIGFsbCB0cmlnZ2VycyBhcmUgbWV0LCBvdGhlcndpc2'
    b'UgcmVjb3JkaW5nIHdpbGwgc3RhcnQgd2hlbiBhbnkgc2Vuc29yIGNvbmRpdGlvbiBpcyBt'
    b'ZXRAGEEKUAWQUHJlc3N1cmUgVHJpZ2dlclABgwUAJFEAgQBQFJlDb25maWdbMHgwMUZGMj'
    b'RdICYgMSA9PSAwQILlUAWVUHJlc3N1cmUgVHJpZ2dlciwgTG93UAGDAwAkUQGDAV+QUCWC'
    b'UGFRIYMB1MBQFbNTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaWRlIHByZXNzdX'
    b'JlIHdpbmRvdy5AguZQBZZQcmVzc3VyZSBUcmlnZ2VyLCBIaWdoUAGDBAAkUQGDAa2wUCWC'
    b'UGFRIYMB1MBQFbNTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaWRlIHByZXNzdX'
    b'JlIHdpbmRvdy5AGEEXUAWTVGVtcGVyYXR1cmUgVHJpZ2dlclABgwUBJFEAgQBQFJlDb25m'
    b'aWdbMHgwMUZGMjRdICYgMiA9PSAwQGPqUAWYVGVtcGVyYXR1cmUgVHJpZ2dlciwgTG93UA'
    b'GDAwEkUQKB9lAlgUNREoHYUSKBUFAVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG91'
    b'dHNpZGUgdGVtcGVyYXR1cmUgd2luZG93LkBj61AFmVRlbXBlcmF0dXJlIFRyaWdnZXIsIE'
    b'hpZ2hQAYMEASRRAoEjUCWBQ1ESgdhRIoFQUBW2U2V0IHRvIHN0YXJ0IHNhbXBsaW5nIHdo'
    b'ZW4gb3V0c2lkZSB0ZW1wZXJhdHVyZSB3aW5kb3cuQBhBolAFmkFjY2VsZXJhdGlvbiBUcm'
    b'lnZ2VyICgxNmcpUAGDB/8gUBXUTG93IHBvd2VyIHRyaWdnZXIgdXNpbmcgZGlnaXRhbCBh'
    b'Y2NlbGVyb21ldGVyIHdpdGggREMgcmVzcG9uc2UuIE5vIHByZS10cmlnZ2VyIGRhdGEuUQ'
    b'CBAFARgwX/CFARgwf/UFAkgFAUlENvbmZpZ1sweDFmZjIwXSA9PSAwQENAsVAFlkFjY2Vs'
    b'ZXJhdGlvbiBUaHJlc2hvbGRQAYME/yBRAYFQUgOIP7AQEBAQEBBRI4hAMAAAAAAAAFETgF'
    b'AV8kxvdyBwb3dlciB0cmlnZ2VyLiBUaGUgbWluaW11bSBhY2NlbGVyYXRpb24gKHBvc2l0'
    b'aXZlIG9yIG5lZ2F0aXZlKSB0byB0cmlnZ2VyIGEgcmVjb3JkaW5nLiBNdXN0IGJlIGdyZW'
    b'F0ZXIgdGhhbiAwLkAnxlABgwX/IFEBgQdBB5FQBY5YIEF4aXMgVHJpZ2dlckEHkVAFjlkg'
    b'QXhpcyBUcmlnZ2VyQQeRUAWOWiBBeGlzIFRyaWdnZXJAGEGiUAWZQWNjZWxlcmF0aW9uIF'
    b'RyaWdnZXIgKDhnKVABgwf/UFAV1ExvdyBwb3dlciB0cmlnZ2VyIHVzaW5nIGRpZ2l0YWwg'
    b'YWNjZWxlcm9tZXRlciB3aXRoIERDIHJlc3BvbnNlLiBObyBwcmUtdHJpZ2dlciBkYXRhLl'
    b'EAgQBQEYMF/whQEYMH/yBQJIBQFJRDb25maWdbMHgxZmY1MF0gPT0gMEBDQLJQBZZBY2Nl'
    b'bGVyYXRpb24gVGhyZXNob2xkUAGDBP9QUQGCIABSA4g/IAAQABAAEFEjiEAgAAAAAAAAUR'
    b'OAUBXyTG93IHBvd2VyIHRyaWdnZXIuIFRoZSBtaW5pbXVtIGFjY2VsZXJhdGlvbiAocG9z'
    b'aXRpdmUgb3IgbmVnYXRpdmUpIHRvIHRyaWdnZXIgYSByZWNvcmRpbmcuIE11c3QgYmUgZ3'
    b'JlYXRlciB0aGFuIDAuQCfGUAGDBf9QUQGBB0EHkVAFjlggQXhpcyBUcmlnZ2VyQQeRUAWO'
    b'WSBBeGlzIFRyaWdnZXJBB5FQBY5aIEF4aXMgVHJpZ2dlckAPgEBPgEAoRoBQBYxNZWFzdX'
    b'JlbWVudHNACEC6UAWbMTZnIERDIEFjY2VsZXJhdGlvbiAoQ2ggMzIpQCefUAGDAf8gUQGB'
    b'AUEHklAFj0VuYWJsZSBBbGwgQXhlc0AX91AFi1NhbXBsZSBSYXRlUAGDAv8gUQGCAZBQJY'
    b'JIelAUk25vdCBDb25maWdbMHgxZmYyMF1BB4VRAYIMgEEHhVEBggZAQQeFUQGCAyBBB4VR'
    b'AYIBkEEHhFEBgchBB4RRAYFkQQeEUQGBMkEHhFEBgRlBB4RRAYEMQAhBG1AFmjhnIERDIE'
    b'FjY2VsZXJhdGlvbiAoQ2ggODApQCefUAGDAf9QUQGBAUEHklAFj0VuYWJsZSBBbGwgQXhl'
    b'c0AXQNhQBYtTYW1wbGUgUmF0ZVABgwL/UFEBggH0UCWCSHpQFd5Mb3cgcGFzcyBmaWx0ZX'
    b'IgaXMgc2V0IHRvIDI1JSBvZiBkYXRhIHJhdGUsIHNvIGEgZGF0YSByYXRlIG9mIDIwMDBI'
    b'eiBoYXMgYSBMUCBmaWx0ZXIgYXQgNTAwIEh6UBSTbm90IENvbmZpZ1sweDFmZjUwXUEHhV'
    b'EBgg+gQQeFUQGCB9BBB4VRAYID6EEHhVEBggH0QQeEUQGB+kEHhFEBgX1BB4RRAYE/QQeE'
    b'UQGBIEEHhFEBgRBACEEwUAWoQ29udHJvbCBQYWQgVGVtcGVyYXR1cmUvUHJlc3N1cmUgKE'
    b'NoIDU5KUAnQJtRAYEDUAGDAf87UBWlRW5hYmxlL0Rpc2FibGUgc2Vuc29ycyBvbiBjb250'
    b'cm9sIHBhZEEHnFAFmUVuYWJsZSBQcmVzc3VyZSAoQ2ggNTkuMClBB59QBZxFbmFibGUgVG'
    b'VtcGVyYXR1cmUgKENoIDU5LjEpQQelUAWiRW5hYmxlIFJlbGF0aXZlIEh1bWlkaXR5IChD'
    b'aCA1OS4yKUAR41AFi1NhbXBsZSBSYXRlUAGDAv87UBSUQ29uZmlnWzB4MWZmM0JdID09ID'
    b'BQFaRTZXQgc2FtcGxlIGZyZXF1ZW5jeSBmcm9tIDEgdG8gMTAgSHpRAYEKUCWCSHpREYEB'
    b'USGBCkAIQc1QBZlJbmVydGlhbCBNZWFzdXJlbWVudCBVbml0QAdAllAFkEFjcXVpc2l0aW'
    b'9uIE1vZGVQAYMB/ytQJYEgUQGBEEEHilAFg09mZlEBgQBBB6hQBaFBYnNvbHV0ZSBPcmll'
    b'bnRhdGlvbiAoUXVhdGVybmlvbilRAYEIQQeoUAWhUmVsYXRpdmUgT3JpZW50YXRpb24gKF'
    b'F1YXRlcm5pb24pUQGBEEEHj1AFiFJvdGF0aW9uUQGBAkABQIZQBZVPcmllbnRhdGlvbiBE'
    b'YXRhIFJhdGVQAYMCCCtRAYFkUCWCSHpREYEBUSGByFAVvVNhbXBsZSByYXRlIGZvciBkaX'
    b'JlY3Rpb25hbCBkYXRhIGZyb20gMSB0byAyMDAgSHouIENoYW5uZWwgNjVQFJRDb25maWdb'
    b'MHgwMUZGMkJdIDwgOEABQIlQBZJSb3RhdGlvbiBEYXRhIFJhdGVQAYMCAitRAYFkUCWCSH'
    b'pREYEBUSGByFAVvlNldCBzYW1wbGUgcmF0ZSBmb3Igcm90YXRpb24gZGF0YSBmcm9tIDEg'
    b'dG8gMjAwIEh6LiBDaGFubmVsIDQ3UBSZQ29uZmlnWzB4MDFGRjJCXSAmIDIgPT0gMEAIQP'
    b'FQBaVJbnRlcm5hbCBUZW1wZXJhdHVyZS9QcmVzc3VyZSAoQ2ggMzYpQCdAxVEBgQNQAYMB'
    b'/yRQFfdFbmFibGUvRGlzYWJsZSBzbG93ZXIgaW50ZXJuYWwgZW52aXJvbm1lbnRhbCBzZW'
    b'5zb3JzLiBOT1RFOiBUZW1wZXJhdHVyZSBpcyByZXF1aXJlZCBmb3IgbWFpbiBhY2NlbGVy'
    b'b21ldGVyIG1lYXN1cmVtZW50c0EHnFAFmUVuYWJsZSBQcmVzc3VyZSAoQ2ggMzYuMClBB5'
    b'9QBZxFbmFibGUgVGVtcGVyYXR1cmUgKENoIDM2LjEpQAhAkFAFoENvbnRyb2wgUGFkIExp'
    b'Z2h0IFNlbnNvciAoQ2ggNzYpQCfqUAGDAf9MUBW+RW5hYmxlL0Rpc2FibGUgbGlnaHQgc2'
    b'Vuc29yIGF0IDQgSHogLSBOT1RFOiBESVNBQkxFRCBGT1IgQUxQSEFBB6BQBZ1FbmFibGUg'
    b'TGlnaHQgU2Vuc29yIChDaCA3Ni4wKUAPgEBPgEoIgEoogEpIgA=='
)



"""
Windows-specific functions; primarily filesystem-related. The parent package
imports the appropriate version for the host OS.
"""

__author__ = "dstokes"
__copyright__ = "Copyright 2023 Mide Technology Corporation"

__all__ = ('deviceChanged', 'getDeviceList', 'getBlockSize', 'getFreeSpace',
           'getDriveInfo', 'readRecorderClock', 'readUncachedFile')

import ctypes
import errno
import logging
import os
from pathlib import Path
import sys
from time import time
from typing import List, Optional, Tuple, Union
import warnings

from .types import Drive, Epoch, Filename

logger = logging.getLogger('endaq.device')

# ==============================================================================
#
# ==============================================================================

if 'win' in sys.platform and sys.platform != 'darwin':
    kernel32 = ctypes.windll.kernel32
    import pywintypes
    import win32api
    import win32con
    import win32file
else:
    kernel32 = None


# ==============================================================================
#
# ==============================================================================

def getDriveInfo(dev: Filename) -> Drive:
    """ Get general device information. Not currently used.
    """
    dev = os.path.realpath(str(dev))

    volumeNameBuffer = ctypes.create_unicode_buffer(1024)
    fileSystemNameBuffer = ctypes.create_unicode_buffer(1024)
    serial_number = ctypes.c_uint(0)
    # file_system_flags =  ctypes.c_uint(0)
    kernel32.GetVolumeInformationW(
        ctypes.c_wchar_p(dev),
        volumeNameBuffer,
        ctypes.sizeof(volumeNameBuffer),
        ctypes.byref(serial_number),
        None,  # max_component_length,
        None,  # ctypes.byref(file_system_flags),
        fileSystemNameBuffer,
        ctypes.sizeof(fileSystemNameBuffer)
    )

    try:
        sn = "%08X" % serial_number.value
    except (AttributeError, TypeError):
        sn = None

    try:
        fs = fileSystemNameBuffer.value
    except AttributeError:
        fs = None

    return Drive(path=dev, label=volumeNameBuffer.value, sn=sn,
                 fs=fs, type=win32file.GetDriveType(dev))
    

def readUncachedFile(filename: Filename) -> bytes:
    """ Read a file, circumventing the disk cache. Returns the data read.
    """
    if isinstance(filename, Path):
        filename = str(filename)

    filename = os.path.realpath(filename)
    root = os.path.dirname(filename)

    if not os.path.isfile(filename):
        raise IOError(errno.ENOENT, 'No such file', filename)

    # For efficiency, this reads entire blocks.
    # Get the file size, rounded up to bytes per filesystem block.
    size = os.path.getsize(filename)
    spc, bps, _fc, _tc  = win32file.GetDiskFreeSpace(root)
    bpc = spc * bps

    if size > bpc:
        blocks = size // bpc
        if bpc % size:
            blocks += 1
        size = blocks * bpc
    else:
        size = bpc

    try:
        f1 = win32file.CreateFile(filename,
                                  win32con.GENERIC_READ, 
                                  win32con.FILE_SHARE_READ,
                                  None,
                                  win32con.OPEN_EXISTING,
                                  win32con.FILE_FLAG_NO_BUFFERING,
                                  0)
        _hr, data = win32file.ReadFile(f1, int(size))
        win32api.CloseHandle(f1)        
        return data
    
    except pywintypes.error as err: 
        raise IOError(err.strerror)


def readRecorderClock(clockFile: Filename,
                      pause: bool = True,
                      timeout: Union[int, float] = 3) -> Tuple[Epoch, Epoch]:
    """ Read a (recorder) clock file, circumventing the disk cache. Returns
        the system time and the encoded device time.

        :param clockFile: The full path to the device's clock file (e.g.,
            `/SYSTEM/DEV/CLOCK`).
        :param pause: If `True` (default), wait until the data in the clock
            file changes (e.g. a new tick/second has started) before
            returning. This is a means to maximize accuracy.
        :param timeout: Seconds to wait for a successful read, when `pause`
            is `True`, before raising a `TimeoutError`.
        :return: The host time, and the unparsed contents of the device clock
            file.
    """
    root = os.path.abspath(os.path.join(os.path.dirname(clockFile), '..', '..'))
    
    try:
        f1 = win32file.CreateFile(clockFile,
                                  win32con.GENERIC_READ, 
                                  win32con.FILE_SHARE_READ,
                                  None,
                                  win32con.OPEN_EXISTING,
                                  win32con.FILE_FLAG_NO_BUFFERING,
                                  0)
    except pywintypes.error as err:
        raise IOError(err.strerror)

    # For efficiency, this reads entire blocks.
    # It is assumed the clock value is smaller than one block.
    spc, bps, _fc, _tc  = win32file.GetDiskFreeSpace(root)
    bpc = spc * bps

    _hr, lastTime = win32file.ReadFile(f1, bpc)
    thisTime = lastTime
    sysTime = time()
    win32file.SetFilePointer(f1, 0, win32file.FILE_BEGIN)

    deadline = sysTime + timeout

    if pause:
        while lastTime == thisTime:
            sysTime = time()
            if timeout > 0 and sysTime > deadline:
                raise TimeoutError('Timed out waiting for CLOCK file to update')
            _hr, thisTime = win32file.ReadFile(f1, bpc)
            sysTime = (time() + sysTime)/2
            win32file.SetFilePointer(f1, 0, win32file.FILE_BEGIN)

    win32api.CloseHandle(f1)
    return sysTime, thisTime

# ==============================================================================
# 
# ==============================================================================


def getDeviceList(types: dict, strict: bool = True) -> List[Drive]:
    """ Get a list of data recorders, as their respective drive letter.
    """
    drivebits = kernel32.GetLogicalDrives()
    result = []
    for letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        try:
            if drivebits & 1:
                driveLetter = '%s:\\' % letter
                devtype = win32file.GetDriveType(driveLetter)
                # First cut: only consider removable drives
                if devtype == win32file.DRIVE_REMOVABLE or not strict:
                    info = getDriveInfo(driveLetter)
                    for t in types:
                        if t.isRecorder(info, strict=strict):
                            result.append(info)
                            break

        except IOError as err:
            # WindowsError 433 is not uncommon for devices that just started recording.
            # Ignore it, warn about any others.
            if getattr(err, 'winerror', None) != 433:
                msg = ("getDeviceList(): Could not access {}:/ ({}); "
                       "ignoring error and continuing".format(letter, err))
                warnings.warn(msg)
                logger.error(msg)

        drivebits >>= 1
    return result


# Module-level globals for caching last discovered logical drives and recorders
_LAST_DEVICES = 0       # Bitmap of logical drives (Z...A)
_LAST_RECORDERS = None  # Tuple of recorder paths (e.g. `["D:\\", "E:\\"]`)


def deviceChanged(recordersOnly: bool, types: Optional[dict] = None, clear: bool = False) -> bool:
    """ Returns `True` if a drive has been connected or disconnected since
        the last call to `deviceChanged()`.
        
        :param recordersOnly: If `False`, any change to the mounted drives
            is reported as a change. If `True`, the mounted drives are checked
            and `True` is only returned if the change occurred to a recorder.
            Checking for recorders only takes marginally more time.
        :param types: A list of known `Recorder` classes to detect. Used if
            `recordersOnly` is `True`.
        :param clear: If `True`, clear the cache of previously-detected
            drives and devices.
    """
    global _LAST_DEVICES, _LAST_RECORDERS
    
    if clear:
        _LAST_DEVICES = 0
        _LAST_RECORDERS = None
    
    newDevices = kernel32.GetLogicalDrives()
    changed = newDevices != _LAST_DEVICES
    _LAST_DEVICES = newDevices
    
    # if not changed or not recordersOnly:
    if not recordersOnly:
        return changed

    newRecorders = tuple(getDeviceList(types=types))
    changed = newRecorders != _LAST_RECORDERS
    _LAST_RECORDERS = newRecorders
    return changed

# ==============================================================================
# 
# ==============================================================================


def getFreeSpace(path: Filename):
    """ Return the free space (in bytes) on a drive.
        
        :param path: The path to the drive to check. Can be a subdirectory.
        :return: The free space on the drive, in bytes.
        :rtype: int
    """
    if isinstance(path, Path):
        path = str(path)

    free_bytes = ctypes.c_ulonglong(0)
    kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(path), None, None, 
                                 ctypes.pointer(free_bytes))
    return free_bytes.value


def getBlockSize(path: Filename):
    """ Return the bytes per sector and sectors per cluster of a drive.

        :param path: The path to the drive to check. Can be a subdirectory.
        :return: A tuple containing the bytes/sector and sectors/cluster.
    """
    if isinstance(path, Path):
        path = str(path)

    sectorsPerCluster = ctypes.c_ulonglong(0)
    bytesPerSector = ctypes.c_ulonglong(0)
    kernel32.GetDiskFreeSpaceW(ctypes.c_wchar_p(path), 
                                 ctypes.pointer(sectorsPerCluster),
                                 ctypes.pointer(bytesPerSector),
                                 None, None)
    
    return bytesPerSector.value, sectorsPerCluster.value



"""
Some basic utility functions, for internal use.
"""

import errno
import os.path
import pathlib
import shutil
import typing

import logging
logger = logging.getLogger("endaq.device")


def makeBackup(filename: typing.Union[str, pathlib.Path]) -> bool:
    """ Create a backup copy of the given file. For use in conjunction with
        `restoreBackup()`.
    """
    try:
        backupFilename = filename + "~"
        if os.path.exists(filename):
            shutil.copy2(filename, backupFilename)
            return True
    except IOError as err:
        logger.error(f'Failed to create backup of {filename} '
                     f'({errno.errorcode.get(err.errno, "?")}: {err.strerror}), ignoring')
    return False


def restoreBackup(filename: typing.Union[str, pathlib.Path],
                  remove: bool = False) -> bool:
    """ Restore a backup copy of a file, overwriting the file. For use in
        conjunction with `makeBackup()`.
    """
    try:
        backupFilename = filename + "~"
        if os.path.exists(backupFilename):
            shutil.copy2(backupFilename, filename)
            if remove:
                os.remove(backupFilename)
            return True
    except IOError as err:
        logger.error(f'Failed to restore backup of {filename} '
                     f'({errno.errorcode.get(err.errno, "?")}: {err.strerror}), ignoring')
    return False


def cleanProps(el: dict):
    """ Recursively remove unknown elements (``"UnknownElement"`` keys) from
        a dictionary of device properties. The original data may contain
        nested dictionaries and lists. For preparing data dumped from EBML
        for re-encoding.

        Nested dictionaries and lists are deep-copied. Note: the contents of
        `bytearray` objects are not duplicated; the copy of the dictionary
        will reference the same ones as the original.

        :return: A deep copy of the original dictionary, minus unknown
            elements.
    """
    if isinstance(el, list):
        return [cleanProps(x) for x in el]
    elif not isinstance(el, dict):
        return el

    return {k: cleanProps(v) for k, v in el.items() if k != "UnknownElement"}



"""
Classes for representing specific models of enDAQ data recoreder.
"""

__author__ = "dstokes"
__copyright__ = "Copyright 2023 Mide Technology Corporation"

import re
from typing import Callable, Optional, Union
import warnings

from .base import Recorder
from .exceptions import ConfigError, UnsupportedFeature


# ==============================================================================
# 
# ==============================================================================

class EndaqS(Recorder):
    """ An enDAQ S-series data recorder from Mide Technology Corporation. 
    """

    SN_FORMAT = "S%07d"
        
    manufacturer = "Midé Technology Corporation"
    homepage = "https://endaq.com/collections/endaq-shock-recorders-vibration-data-logger-sensors"

    _NAME_PATTERN = re.compile(r'(^S(\d|\d\d)-.*)|(^SF-DR4-(02|04))')


# ==============================================================================
# 
# ==============================================================================

class EndaqW(EndaqS):
    """ An enDAQ W-series wireless-enabled data recorder from Mide Technology Corporation.
    """
    SN_FORMAT = "W%07d"

    # Part number starts with "W", a 1-2 digit number, and "-"
    _NAME_PATTERN = re.compile(r'(^W(\d|\d\d)-.*)|(^SF-DR4-(01|03))')


    def setAP(self, ssid: str, password: Optional[str] = None):
        """ Quickly set the Wi-Fi access point (router) and password.

            :param ssid: The SSID (name) of the wireless access point.
            :param password: The access point password.
        """
        # FUTURE: Remove EndaqW.setAP()
        warnings.warn("Direct control moved to `command` attribute; use "
                      "recorder.command.setAP()", DeprecationWarning)

        if self.isVirtual:
            raise ConfigError('Cannot configure a virtual device')
        if not self.command:
            raise UnsupportedFeature('Device has no command interface')

        return self.command.setAP(ssid, password)


    def setWifi(self,
                wifi_data: dict,
                timeout: int = 10,
                interval: float = 1.25,
                callback: Optional[Callable] = None):
        """ Configure all known Wi-Fi access points. Applicable only
            to devices with Wi-Fi hardware. The data is in the form of a
            list of dictionaries with the following keys:

            * ``"SSID"``: The Wi-Fi access point name (string)
            * ``"Password"``: The access point's password (string, optional)
            * ``"Selected"``: 1 if the device should use this AP, 0 if not

            Note that devices may not support configuring multiple Wi-Fi AP.
            In most cases, only one should be specified, and it should be
            marked as selected.

            Note: This method is deprecated. Use `recorder.command.setWiFi()`
            instead.

            :param wifi_data: The information about the Wi-Fi networks to be
                set on the device.
            :param timeout: Time (in seconds) to wait for a response before
                raising a :class:`~.endaq.device.DeviceTimeout` exception.
            :param interval: Time (in seconds) between checks for a response.
            :param callback: A function to call each response-checking cycle.
                If the callback returns `True`, the wait for a response will be
                cancelled. The callback function should take no arguments.
        """
        # FUTURE: Remove EndaqW.setWiFi()
        warnings.warn("Direct control moved to `command` attribute; use "
                      "recorder.command.setWiFi()", DeprecationWarning)

        if self.isVirtual:
            raise ConfigError('Cannot configure a virtual device')
        if not self.command:
            raise UnsupportedFeature('Device has no command interface')

        if not wifi_data:
            # For legacy code. CommandInterface version does not warn.
            warnings.warn("Use command.queryWifi() to get Wi-Fi status, "
                          "not setWifi(None)", DeprecationWarning)
            return self.command.queryWifi(timeout, interval / 5, callback)

        return self.command.setWifi(wifi_data=wifi_data,
                                    timeout=timeout,
                                    interval=interval,
                                    callback=callback)


    def queryWifi(self,
                  timeout: int = 10,
                  interval: float = .25,
                  callback: Optional[Callable] = None) -> Union[None, dict]:
        """ Check the current state of the Wi-Fi (if present).

            Note: This method is deprecated. Use `recorder.command.queryWiFi()`
            instead.

            :param timeout: Time (in seconds) to wait for a response before
                raising a `DeviceTimeout` exception.
            :param interval: Time (in seconds) between checks for a response.
            :param callback: A function to call each response-checking cycle.
                If the callback returns `True`, the wait for a response will be
                cancelled. The callback function should require no arguments.
            :return: None if no information was recieved, else it will return
                the information from the ``QueryWiFiResponse`` command (this
                return statement is not used anywhere)
        """
        # FUTURE: Remove EndaqW.queryWifi()
        warnings.warn("Direct control moved to `command` attribute; use "
                      "recorder.command.queryWifi()", DeprecationWarning)

        if self.isVirtual:
            raise ConfigError('Cannot configure a virtual device')
        if not self.command:
            raise UnsupportedFeature('Device has no command interface')

        return self.command.queryWifi(timeout=timeout, interval=interval, callback=callback)


    def scanWifi(self, timeout: int = 10,
                 interval: float = .25,
                 callback: Optional[Callable] = None) -> Union[None, list]:
        """ Initiate a scan for Wi-Fi access points (APs). Applicable only
            to devices with Wi-Fi hardware.

            The resluts are returned as a list of dictionaries, one for each
            access point, with keys:

            * ``SSID`` (str): The access point name.
            * ``RSSI`` (int): The AP's signal strength.
            * ``AuthType`` (int): The authentication (security) type.
              Currently, this is either 0 (no authentication) or 1
              (any authentication).
            * ``Known`` (bool): Is this access point known (i.e. has
              a stored password on the device)?
            * ``Selected`` (bool): Is this the currently selected AP?

            Note: This method is deprecated. Use `recorder.command.scanWiFi()`
            instead.

            :param timeout: Time (in seconds) to wait for a response before
                raising a :class:`~.endaq.device.DeviceTimeout` exception.
            :param interval: Time (in seconds) between checks for a response.
            :param callback: A function to call each response-checking cycle.
                If the callback returns `True`, the wait for a response will
                be cancelled. The callback function should take no arguments.
            :return: A list of dictionaries, described above.
        """
        # FUTURE: Remove EndaqW.scanWifi()
        warnings.warn("Direct control moved to `command` attribute; use "
                      "recorder.command.scanWifi()", DeprecationWarning)

        if self.isVirtual:
            raise ConfigError('Cannot configure a virtual device')
        elif not self.command:
            raise UnsupportedFeature('Device has no command interface')

        return self.command.scanWifi(timeout=timeout, interval=interval, callback=callback)


    def updateESP32(self,
                    firmware: str,
                    timeout: float = 10,
                    callback: Optional[Callable] = None):
        """ Update the ESP32 firmware.

            Note: This method is deprecated. Use `recorder.command.updateESP32()`
            instead.

            :param firmware: The name of the ESP32 firmware package (.bin).
            :param timeout: Time (in seconds) to wait for the recorder to
                respond. 0 will return immediately.
            :param callback: A function to call each response-checking
                cycle. If the callback returns `True`, the wait for a response
                will be cancelled. The callback function should take no
                arguments.
            :return:
        """
        # FUTURE: Remove EndaqW.updateESP32()
        warnings.warn("Direct control moved to `command` attribute; use "
                      "recorder.command.updateESP32()", DeprecationWarning)

        return self.command.updateESP32(firmware=firmware,
                                        timeout=timeout,
                                        callback=callback)



"""
Default ConfigUI for S5-R500D40 data recorders 
"""

from base64 import b64decode

DEFAULT_CONFIG_UI = b64decode(
    b'd3dW8kAoRjBQBYdHZW5lcmFsQAXzUAWLRGV2aWNlIE5hbWVQAYMI/39RJIFAUBXYQSBjdX'
    b'N0b20gbmFtZSBmb3IgdGhlIHJlY29yZGVyLiBOb3QgdGhlIHNhbWUgYXMgdGhlIHZvbHVt'
    b'ZSBsYWJlbC4gNjQgY2hhcmFjdGVycyBtYXhpbXVtLkAF81AFjERldmljZSBOb3Rlc1ABgw'
    b'n/f1EkggEAUBXSQ3VzdG9tIG5vdGVzIGFib3V0IHRoZSByZWNvcmRlciAocG9zaXRpb24s'
    b'IHVzZXIgSUQsIGV0Yy4pLiAyNTYgY2hhcmFjdGVycyBtYXhpbXVtLlE0gQNABUCZUAWVQ3'
    b'VzdG9tIFJlY29yZGluZyBUYWdzUAGDF/9/UQSAUSSBQFAV8UtleXdvcmRzIGZvciBvcmdh'
    b'bml6aW5nIGFuZCBpZGVudGlmeWluZyByZWNvcmRpbmdzLCBzZXBhcmF0ZWQgYnkgY29tbW'
    b'Egb3Igc3BhY2UuIExpbWl0IGlzIDIwIHRhZ3MsIDY0IGNoYXJhY3RlcnMuQAX7UAWTUmVj'
    b'b3JkaW5nIERpcmVjdG9yeVABgxT/f1EEhlJFQ09SRFEkgRBQFJNDb25maWdbMHgxNkZGN0'
    b'ZdPT0wUBW5RGlyZWN0b3J5IGZvciBzZXJpYWxpemVkIHJlY29yZGluZ3MuIDE2IGNoYXJh'
    b'Y3RlciBtYXhpbXVtQAVAo1AFlVJlY29yZGluZyBGaWxlIFByZWZpeFABgxX/f1EkgRBQFe'
    b'hQcmVmaXggZm9yIHNlcmlhbGl6ZWQgcmVjb3JkaW5nIG5hbWVzLiBJZiBibGFuaywgZGV2'
    b'aWNlIHNlcmlhbCBudW1iZXIgd2lsbCBiZSB1c2VkLiAxNiBjaGFyYWN0ZXIgbWF4aW11bV'
    b'AUk0NvbmZpZ1sweDE2RkY3Rl09PTBAB0E8UAWPRmlsZSBOYW1lIFN0eWxlUAGDFv9/UQGB'
    b'AVAVQPxUaW1lLUJhc2VkIGZpbGUgbmFtZXMgdXNlIGEgZGlyZWN0b3J5IGJhc2VkIG9uIH'
    b'RoZSBkYXRlLCBhbmQgZmlsZSBuYW1lIGJhc2VkIG9uIHRoZSBzZWNvbmRzIHNpbmNlIG1p'
    b'ZG5pZ2h0LgpTZXJpYWxpemVkIGZpbGUgbmFtZXMgdXNlIHRoZSBzcGVjaWZpZWQgZGlyZW'
    b'N0b3J5IGFuZCBmaWxlIG5hbWUgcHJlZml4LCBmb2xsb3dlZCBieSBhIG51bWJlci4gCk51'
    b'bWJlciBpcyByZXNldCB3aGVuIGRpcmVjdG9yeSBvciBwcmVmaXggY2hhbmdlcy5BB41QBY'
    b'pUaW1lIEJhc2VkQQeNUAWKU2VyaWFsaXplZEAHQKlQBYtCdXR0b24gTW9kZVABgxD/f1EB'
    b'gQBQFb9PbmUgU2Vjb25kIFByZXNzIHdpbGwgZGlzYWJsZSAzLXNlY29uZCBwcmVzcyBmb3'
    b'IgYmF0dGVyeSBjaGVjay5BB5dQBZRJbW1lZGlhdGUgU3RhcnQvU3RvcEEHolAFn09uZSBT'
    b'ZWNvbmQgUHJlc3MgZm9yIFN0YXJ0L1N0b3BBB41QBYpTdGFydC1Pbmx5QBdBFFAFjlBsdW'
    b'ctSW4gQWN0aW9uUAGDCv9/UQGBAVAVrVNldCBhY3Rpb24gZm9yIGRldmljZSB3aGVuIHBs'
    b'dWdnZWQgaW4gdG8gVVNCLkEHtVAFskltbWVkaWF0ZWx5IHN0b3AgcmVjb3JkaW5nIGFuZC'
    b'BhcHBlYXIgYXMgVVNCIGRyaXZlQQezUAWwQ29tcGxldGUgcmVjb3JkaW5nIGJlZm9yZSBh'
    b'cHBlYXJpbmcgYXMgVVNCIGRyaXZlQQewUAWtSWdub3JlOiBzdG9wIHJlY29yZGluZyB3aG'
    b'VuIGJ1dHRvbiBpcyBwcmVzc2VkQQelUAWiU3RhcnQgcmVjb3JkaW5nIHdoZW4gVVNCIGNv'
    b'bm5lY3RlZEAj51ABgwv/f1AV3kVudGVyIGxvY2FsIHRpbWV6b25lJ3Mgb2Zmc2V0IGZyb2'
    b'0gVVRDIHRpbWUuIEltcGFjdHMgdGltZXN0YW1wIGluZm9ybWF0aW9uIG9uIHJlY29yZGlu'
    b'ZyBmaWxlcy5AL4BAD4BAT4BAKEjoUAWIVHJpZ2dlcnNAB0DaUAWMVHJpZ2dlciBNb2RlUA'
    b'GDEv9/UQGBAFAVQI9EZWxheSBUaGVuIFRyaWdnZXI6IFdhaXQgc3BlY2lmaWVkIHRpbWUs'
    b'IHRoZW4gd2FpdCBmb3IgdHJpZ2dlcgpEZWxheSBPciBUcmlnZ2VyOiBBY3F1aXJlIGFmdG'
    b'VyIHdhaXQgb3Igb24gdHJpZ2dlci4gQmF0dGVyeSBsaWZlIG1heSBiZSByZWR1Y2VkLkEH'
    b'lVAFkkRlbGF5IFRoZW4gVHJpZ2dlckEHk1AFkERlbGF5IE9yIFRyaWdnZXJAMvRQBY1TdG'
    b'FydCBhdCBUaW1lUAGDD/9/UBXbV2FpdCB1bnRpbCBzZXQgdGltZSBiZWZvcmUgZXZhbHVh'
    b'dGluZyBvdGhlciB0cmlnZ2VyIGNvbmRpdGlvbnMsIGluY2x1ZGluZyBhIHJlY29yZGluZy'
    b'BkZWxheUARrlAFlFJlY29yZGluZyBUaW1lIExpbWl0UAGDDf9/USGE////8FAlh1NlY29u'
    b'ZHNAEUCQUAWYUmVjb3JkaW5nIERlbGF5L0ludGVydmFsUAGDDP9/UQGBAFAlh1NlY29uZH'
    b'NRIYMBUYBQFdhUaW1lIGJlZm9yZSBzdGFydGluZyBhIHJlY29yZGluZywgYW5kIHRpbWUg'
    b'YmV0d2VlbiByZWNvcmRpbmdzIGlmICdSZXRyaWdnZXInIGlzIGNoZWNrZWQuQBGvUAWZUm'
    b'Vjb3JkaW5nIEZpbGUgU2l6ZSBMaW1pdFABgxH/f1EhhP////BQJYNLaUJAEECXUAWJUmV0'
    b'cmlnZ2VyUAGDDv9/UBSwQ29uZmlnWzB4REZGN0ZdPT1udWxsIGFuZCBDb25maWdbMHgxMU'
    b'ZGN0ZdPT1udWxsUBXPV2hlbiBzZXQsIFNsYW0gU3RpY2sgd2lsbCByZS1hcm0gYWZ0ZXIg'
    b'cmVjb3JkaW5nIHRpbWUgb3Igc2l6ZSBsaW1pdCBpcyByZWFjaGVkLkAQQKhQBZ5XYWl0IG'
    b'ZvciBBbGwgU2Vuc29yIENvbmRpdGlvbnNQAYMT/39QFf5XaGVuIHNldCwgcmVjb3JkaW5n'
    b'IHdpbGwgb25seSBzdGFydCB3aGVuIGFsbCB0cmlnZ2VycyBhcmUgbWV0LCBvdGhlcndpc2'
    b'UgcmVjb3JkaW5nIHdpbGwgc3RhcnQgd2hlbiBhbnkgc2Vuc29yIGNvbmRpdGlvbiBpcyBt'
    b'ZXRAGEDuUAWQUHJlc3N1cmUgVHJpZ2dlclABgwUAJFEAgQBAguVQBZVQcmVzc3VyZSBUcm'
    b'lnZ2VyLCBMb3dQAYMDACRRAYMBX5BQJYJQYVEhgwHUwFAVs1NldCB0byBzdGFydCBzYW1w'
    b'bGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cmUgd2luZG93LkCC5lAFllByZXNzdXJlIFRyaW'
    b'dnZXIsIEhpZ2hQAYMEACRRAYMBrbBQJYJQYVEhgwHUwFAVs1NldCB0byBzdGFydCBzYW1w'
    b'bGluZyB3aGVuIG91dHNpZGUgcHJlc3N1cmUgd2luZG93LkAYQPtQBZNUZW1wZXJhdHVyZS'
    b'BUcmlnZ2VyUAGDBQEkUQCBAEBj6lAFmFRlbXBlcmF0dXJlIFRyaWdnZXIsIExvd1ABgwMB'
    b'JFECgfZQJYFDURKB2FEigVBQFbZTZXQgdG8gc3RhcnQgc2FtcGxpbmcgd2hlbiBvdXRzaW'
    b'RlIHRlbXBlcmF0dXJlIHdpbmRvdy5AY+tQBZlUZW1wZXJhdHVyZSBUcmlnZ2VyLCBIaWdo'
    b'UAGDBAEkUQKBI1AlgUNREoHYUSKBUFAVtlNldCB0byBzdGFydCBzYW1wbGluZyB3aGVuIG'
    b'91dHNpZGUgdGVtcGVyYXR1cmUgd2luZG93LkAYQahQBZtBY2NlbGVyYXRpb24gVHJpZ2dl'
    b'ciAoNTAwZylQAYMF/whQEYMH/yBRAIEAUCSAUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAwQE'
    b'NAq1ETiMB/QAAAAAAAUSOIQH9AAAAAAABQBZlBY2NlbGVyYXRpb24gVHJpZ2dlciwgTG93'
    b'UAGDA/8IUQKC+ZpQFeBIaWdoIHBvd2VyIHRyaWdnZXIgd2l0aCBwcmUtdHJpZ2dlciBkYX'
    b'RhLiBOb3QgcmVjb21tZW5kZWQgZm9yIGxvbmcgd2FpdHMgd2l0aG91dCBleHRlcm5hbCBw'
    b'b3dlci5SA4g/jzx2Jp2gJUBDQKxRE4jAf0AAAAAAAFEjiEB/QAAAAAAAUAWaQWNjZWxlcm'
    b'F0aW9uIFRyaWdnZXIsIEhpZ2hQAYME/whRAoIGZlAV4EhpZ2ggcG93ZXIgdHJpZ2dlciB3'
    b'aXRoIHByZS10cmlnZ2VyIGRhdGEuIE5vdCByZWNvbW1lbmRlZCBmb3IgbG9uZyB3YWl0cy'
    b'B3aXRob3V0IGV4dGVybmFsIHBvd2VyLlIDiD+PPHYmnaAlQBhBo1AFmkFjY2VsZXJhdGlv'
    b'biBUcmlnZ2VyICg0MGcpUAGDB/9QUBXUTG93IHBvd2VyIHRyaWdnZXIgdXNpbmcgZGlnaX'
    b'RhbCBhY2NlbGVyb21ldGVyIHdpdGggREMgcmVzcG9uc2UuIE5vIHByZS10cmlnZ2VyIGRh'
    b'dGEuUQCBAFARgwX/CFARgwf/IFAkgFAUlENvbmZpZ1sweDFmZjUwXSA9PSAwQENAslAFlk'
    b'FjY2VsZXJhdGlvbiBUaHJlc2hvbGRQAYME/1BRAYIgAFIDiD9EABQAFAAUUSOIQEQAAAAA'
    b'AABRE4BQFfJMb3cgcG93ZXIgdHJpZ2dlci4gVGhlIG1pbmltdW0gYWNjZWxlcmF0aW9uIC'
    b'hwb3NpdGl2ZSBvciBuZWdhdGl2ZSkgdG8gdHJpZ2dlciBhIHJlY29yZGluZy4gTXVzdCBi'
    b'ZSBncmVhdGVyIHRoYW4gMC5AJ8ZQAYMF/1BRAYEHQQeRUAWOWCBBeGlzIFRyaWdnZXJBB5'
    b'FQBY5ZIEF4aXMgVHJpZ2dlckEHkVAFjlogQXhpcyBUcmlnZ2VyQA+AQE+AQChHxVAFjE1l'
    b'YXN1cmVtZW50c0AIQj1QBZg1MDBnIEFjY2VsZXJhdGlvbiAoQ2ggOClQJIBAJ/5RAYEHUA'
    b'GDAf8IUBWgRW5hYmxlL0Rpc2FibGUgYWNjZWxlcmF0aW9uIGF4ZXNBB5hQBZFFbmFibGUg'
    b'WCAoQ2ggOC4wKVEBgQBBB5hQBZFFbmFibGUgWSAoQ2ggOC4xKVEBgQFBB5hQBZFFbmFibG'
    b'UgWiAoQ2ggOC4yKVEBgQJAEUCaUAWLU2FtcGxlIFJhdGVQAYMC/whQFdhTZXQgbWFpbiBh'
    b'Y2NlbGVyYXRpb24gc2FtcGxlIHJhdGUgZnJvbSAxMCB0byAyMDAwMCBIei4KSGlnaGVyIH'
    b'JhdGUgcmVkdWNlcyBiYXR0ZXJ5IGxpZmUuUBSVQ29uZmlnWzB4MDFGRjA4XSA9PSAwUQGC'
    b'E4hQJYJIelERgQpRIYJOIEARQPxQBaNPdmVycmlkZSBBbnRpYWxpYXNpbmcgRmlsdGVyIE'
    b'N1dG9mZlABgwj/CFEBggPoUCWCSHpRIYJhqFAVQKVEZWZhdWx0LCB3aGVuIG5vdCBjaGVj'
    b'a2VkLCBvZiBvbmUtZmlmdGggc2FtcGxlIGZyZXF1ZW5jeS4gQ3V0LW9mZiBmcmVxdWVuY3'
    b'kgaGFzIDI5JSBhdHRlbnVhdGlvbi4KNXRoIG9yZGVyIEJ1dHRlcndvcnRoLiBTb21lIGF0'
    b'dGVudWF0aW9uIGF0IDYwJSBvZiBjdXQtb2ZmIGZyZXF1ZW5jeS5QFJVDb25maWdbMHgwMU'
    b'ZGMDhdID09IDBACEEcUAWbNDBnIERDIEFjY2VsZXJhdGlvbiAoQ2ggODApQCefUAGDAf9Q'
    b'UQGBAUEHklAFj0VuYWJsZSBBbGwgQXhlc0AXQNhQBYtTYW1wbGUgUmF0ZVABgwL/UFEBgg'
    b'H0UCWCSHpQFd5Mb3cgcGFzcyBmaWx0ZXIgaXMgc2V0IHRvIDI1JSBvZiBkYXRhIHJhdGUs'
    b'IHNvIGEgZGF0YSByYXRlIG9mIDIwMDBIeiBoYXMgYSBMUCBmaWx0ZXIgYXQgNTAwIEh6UB'
    b'STbm90IENvbmZpZ1sweDFmZjUwXUEHhVEBgg+gQQeFUQGCB9BBB4VRAYID6EEHhVEBggH0'
    b'QQeEUQGB+kEHhFEBgX1BB4RRAYE/QQeEUQGBIEEHhFEBgRBACEEwUAWoQ29udHJvbCBQYW'
    b'QgVGVtcGVyYXR1cmUvUHJlc3N1cmUgKENoIDU5KUAnQJtRAYEDUAGDAf87UBWlRW5hYmxl'
    b'L0Rpc2FibGUgc2Vuc29ycyBvbiBjb250cm9sIHBhZEEHnFAFmUVuYWJsZSBQcmVzc3VyZS'
    b'AoQ2ggNTkuMClBB59QBZxFbmFibGUgVGVtcGVyYXR1cmUgKENoIDU5LjEpQQelUAWiRW5h'
    b'YmxlIFJlbGF0aXZlIEh1bWlkaXR5IChDaCA1OS4yKUAR41AFi1NhbXBsZSBSYXRlUAGDAv'
    b'87UBSUQ29uZmlnWzB4MWZmM0JdID09IDBQFaRTZXQgc2FtcGxlIGZyZXF1ZW5jeSBmcm9t'
    b'IDEgdG8gMTAgSHpRAYEKUCWCSHpREYEBUSGBCkAIQc1QBZlJbmVydGlhbCBNZWFzdXJlbW'
    b'VudCBVbml0QAdAllAFkEFjcXVpc2l0aW9uIE1vZGVQAYMB/ytQJYEgUQGBEEEHilAFg09m'
    b'ZlEBgQBBB6hQBaFBYnNvbHV0ZSBPcmllbnRhdGlvbiAoUXVhdGVybmlvbilRAYEIQQeoUA'
    b'WhUmVsYXRpdmUgT3JpZW50YXRpb24gKFF1YXRlcm5pb24pUQGBEEEHj1AFiFJvdGF0aW9u'
    b'UQGBAkABQIZQBZVPcmllbnRhdGlvbiBEYXRhIFJhdGVQAYMCCCtRAYFkUCWCSHpREYEBUS'
    b'GByFAVvVNhbXBsZSByYXRlIGZvciBkaXJlY3Rpb25hbCBkYXRhIGZyb20gMSB0byAyMDAg'
    b'SHouIENoYW5uZWwgNjVQFJRDb25maWdbMHgwMUZGMkJdIDwgOEABQIlQBZJSb3RhdGlvbi'
    b'BEYXRhIFJhdGVQAYMCAitRAYFkUCWCSHpREYEBUSGByFAVvlNldCBzYW1wbGUgcmF0ZSBm'
    b'b3Igcm90YXRpb24gZGF0YSBmcm9tIDEgdG8gMjAwIEh6LiBDaGFubmVsIDQ3UBSZQ29uZm'
    b'lnWzB4MDFGRjJCXSAmIDIgPT0gMEAIQJBQBaBDb250cm9sIFBhZCBMaWdodCBTZW5zb3Ig'
    b'KENoIDc2KUAn6lABgwH/TFAVvkVuYWJsZS9EaXNhYmxlIGxpZ2h0IHNlbnNvciBhdCA0IE'
    b'h6IC0gTk9URTogRElTQUJMRUQgRk9SIEFMUEhBQQegUAWdRW5hYmxlIExpZ2h0IFNlbnNv'
    b'ciAoQ2ggNzYuMClACECyUAWTQk1HMjUwIEd5cm8gKENoIDg0KUAnn1ABgwH/VFEBgQBBB5'
    b'JQBY9FbmFibGUgQWxsIEF4ZXNAF/dQBYtTYW1wbGUgUmF0ZVABgwL/VFEBggGQUCWCSHpQ'
    b'FJNub3QgQ29uZmlnWzB4MWZmNTRdQQeFUQGCDIBBB4VRAYIGQEEHhVEBggMgQQeFUQGCAZ'
    b'BBB4RRAYHIQQeEUQGBZEEHhFEBgTJBB4RRAYEZQQeEUQGBDEAPgEBPgEoIgEoogEpIgA=='
)



"""
Configuration interface tests.
"""

import os.path
import shutil
import pytest

import endaq.device
from endaq.device import getRecorder
from endaq.device import config, ui_defaults
from endaq.device.exceptions import ConfigError

from endaq.device.util import makeBackup

from .fake_recorders import RECORDER_PATHS

# Clear any cached devices, just to be safe
endaq.device.RECORDERS.clear()

# Create parameters, mainly to provide an ID, making the results readable
DEVICES = [pytest.param(getRecorder(path, strict=False), id=os.path.basename(path))
           for path in RECORDER_PATHS]

RECORDER_ROOTS = [os.path.basename(p) for p in RECORDER_PATHS]


@pytest.fixture(scope="session")
def dev_copy_dir(tmpdir_factory):
    """ Fixture to make copies of the fake recorders, for tests that affect
        files on the device (configuration, commands, etc.).
    """
    workdir = tmpdir_factory.mktemp("recorders")
    for src in RECORDER_PATHS:
        dst = workdir / os.path.basename(src)
        shutil.copytree(src, dst)
    return workdir


@pytest.mark.parametrize("dev", DEVICES)
def test_config_basics(dev):
    """ Initial 'sanity test' to verify `ConfigInterface` instances are
        being instantiated.
    """
    assert isinstance(dev.config, config.ConfigInterface)


@pytest.mark.parametrize("dev", DEVICES)
def test_configui_defaults(dev):
    """ Confirm there is default ConfigUI data for each fake recorder.
    """
    assert ui_defaults.getDefaultConfigUI(dev) is not None


@pytest.mark.parametrize("path", RECORDER_ROOTS)
def test_temp(path, dev_copy_dir):
    """ Sanity check that temp copies of devices work.
    """
    endaq.device.RECORDERS.clear()
    dev = getRecorder(dev_copy_dir / path, strict=False)
    assert dev


@pytest.mark.parametrize("path", RECORDER_ROOTS)
def test_config_enable(path, dev_copy_dir):
    """ Test setting, writing, and reading channel enable. These items can
        be handled differently internally, so several channels are tested.
    """
    dev = getRecorder(dev_copy_dir / path, strict=False)

    # Remove existing config.cfg files (if any)
    makeBackup(dev.configFile)
    if os.path.exists(dev.configFile):
        os.remove(dev.configFile)

    # This will instantiate the ConfigInterface
    config = dev.config
    assert config

    # Test channels 8 and 59. These have 3 or more subchannels which can
    # be enabled/disabled individually. Most fake recorders have these.
    for chId in (8, 59):
        if chId not in dev.channels:
            continue

        dev.config.enableChannel(dev.channels[chId][0], False)
        dev.config.enableChannel(dev.channels[chId][1], True)
        dev.config.enableChannel(dev.channels[chId][2], False)

        # Verify these subchannels can only be enabled individually
        with pytest.raises(ConfigError):
            dev.config.enableChannel(dev.channels[chId], True)

        assert dev.config.isEnabled(dev.channels[chId][0]) is False
        assert dev.config.isEnabled(dev.channels[chId][1]) is True
        assert dev.config.isEnabled(dev.channels[chId][2]) is False

    # Test channel 80. It as 3 subchannels, but they cannot be individually
    # enabled/disabled. '*D40' recorders have this channel.
    if 80 in dev.channels:
        dev.config.enableChannel(dev.channels[80], False)
        assert dev.config.isEnabled(dev.channels[80]) is False

        # Verify only the channel as a whole can be enabled
        with pytest.raises(ConfigError):
            dev.config.enableChannel(dev.channels[80][0], True)

    # Save config
    dev.config.applyConfig()

    # Remove the old ConfigInterface and make a new one, ensuring all
    # config values are being read from the config file.
    dev.config = None
    assert dev.config is not config

    for chId in (8, 59):
        if chId not in dev.channels:
            continue

        assert dev.config.isEnabled(dev.channels[chId][0]) is False
        assert dev.config.isEnabled(dev.channels[chId][1]) is True
        assert dev.config.isEnabled(dev.channels[chId][2]) is False

    if 80 in dev.channels:
        assert dev.config.isEnabled(dev.channels[80]) is False
